# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from unittest import mock
import inspect
import sys

# Third Party
import pytest

# Local
from caikit.core import MODEL_MANAGER
from caikit.core.data_model.base import DataBase
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.utils.import_util import (
    UnifiedDataModel,
    clean_lib_names,
    get_data_model,
    get_dynamic_module,
)
from tests.conftest import temp_config

## Setup #########################################################################

multi_lib_names = "caikit_libraryfoo[foo.bar]>=1.0.0-rc15 caikit_librarybar foo-bar==1"
multi_lib_cleaned_list = ["caikit_libraryfoo", "caikit_librarybar", "foo_bar"]

## Tests ########################################################################


### clean_lib_names #############################################################
def test_multi_lib_name_cleaning():
    name_list = clean_lib_names(multi_lib_names)
    assert len(name_list) == len(multi_lib_cleaned_list)
    assert multi_lib_cleaned_list == name_list


def test_single_lib_name_cleaning():
    name_list = clean_lib_names("caikit_libraryfoo[foo.bar]>=1.0.0-rc15")
    assert name_list[0] == "caikit_libraryfoo"


def test_single_lib_name_no_option_dep():
    name_list = clean_lib_names("caikit_libraryfoo>=1.0.0-rc15")
    assert name_list[0] == "caikit_libraryfoo"


def test_single_lib_name_no_version():
    name_list = clean_lib_names("caikit_libraryfoo[foo.bar]")
    assert name_list[0] == "caikit_libraryfoo"


### get_dynamic_module #############################################################


def test_get_caikit_library_throws_on_nonimportable_lib():
    """If an invalid module is provided to caikit lib setup, it throws a ValueError"""
    with pytest.raises(CaikitRuntimeException):
        get_dynamic_module("caikit_bad_lib")


def test_get_caikit_library_loads_sample_stdlib_module():
    """If a valid module is provided to caikit lib setup, that module is returned"""
    # Standard
    import sys

    sample_module = get_dynamic_module("sys")
    assert sample_module == sys


def test_get_caikit_library_loads_caikit_core():
    """If caikit.core is provided to caikit lib setup, the module is returned"""
    # Local
    import caikit.core

    sample_module = get_dynamic_module("caikit.core")
    assert sample_module == caikit.core


### get_data_model #############################################################


def test_get_data_model_throws_on_nonimportable_lib():
    """If an invalid module is provided to get_data_model, it throws a ValueError"""
    with temp_config({"runtime": {"library": "caikit_bad_lib"}}):
        with pytest.raises(CaikitRuntimeException):
            get_data_model()


def test_get_data_model_ok_on_lib_with_no_data_model():
    """If a valid module with no data model is provided to get_data_model it
    returns an empty module object
    """
    with temp_config({"runtime": {"library": "sys"}}):
        data_model = get_data_model()
        assert isinstance(data_model, UnifiedDataModel)


def test_get_data_model_is_accessible():
    """get_data_model should return the stuff in the `sample_lib.data_model` package"""
    # Local
    import sample_lib

    cdm = get_data_model()
    attrs_match = [
        (hasattr(cdm, attr_name) and getattr(cdm, attr_name) == attr_val)
        for attr_name, attr_val in vars(sample_lib.data_model).items()
        if not attr_name.startswith("_")
        and inspect.isclass(attr_val)
        and issubclass(attr_val, DataBase)
    ]
    assert all(attrs_match)


def test_multiple_caikit_libraries():
    """Make sure that multiple libraries can be imported and merged into a
    unified data model
    """
    lib_names = ["caikit.interfaces.runtime", "sample_lib"]
    with temp_config({"runtime": {"library": " ".join(lib_names)}}):
        cdm = get_data_model()
        for lib_name in lib_names:
            assert lib_name in sys.modules
            lib_mod = sys.modules[lib_name]
            attrs_match = [
                (hasattr(cdm, attr_name) and getattr(cdm, attr_name) == attr_val)
                for attr_name, attr_val in vars(lib_mod.data_model).items()
                if not attr_name.startswith("_")
                and inspect.isclass(attr_val)
                and issubclass(attr_val, DataBase)
            ]
            assert all(attrs_match)


def test_get_data_model_init_components_after_import():
    """Make sure that the model_manager's init_components is called _after_ the
    dynamic import of the runtime library
    """
    with temp_config({"runtime": {"library": "sample_lib"}}):
        call_mock = mock.MagicMock()
        with mock.patch.object(
            MODEL_MANAGER,
            "initialize_components",
            call_mock,
        ):
            with mock.patch(
                "caikit.runtime.utils.import_util._get_cdm_from_lib", call_mock
            ):
                get_data_model()
                # Make sure the mock was called for both _get_cdm_from_lib and
                # initialize_components
                assert len(call_mock.call_args_list) > 1

                # Make sure the last call is the only one with no args
                # indicating that it's the call to initialize_components
                assert not call_mock.call_args_list[-1].args
                assert all(call.args for call in call_mock.call_args_list[:-1])



# Standard
import os
import shutil

# First Party
import alog

# Local
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException

log = alog.use_channel("TEST-FIXTURE")


class Fixtures:
    """
    This class contains test fixtures that can be used to
    perform model manager operations with input models for tests
    """

    @staticmethod
    def load_model(model_id, local_model_path, model_type):
        """Load a model using model manager load model implementation"""
        try:
            model_manager = ModelManager.get_instance()
            model_manager.load_model(model_id, local_model_path, model_type)
        except CaikitRuntimeException as e:
            log.error({"message": e.message, "model_id": model_id, "error_id": e.id})
            raise e

    @staticmethod
    def unload_all_models():
        """Unload all loaded models using model manager unload models implementation"""
        try:
            model_manager = ModelManager.get_instance()
            model_manager.unload_all_models()
        except CaikitRuntimeException as e:
            log.error({"message": e.message, "error_id": e.id})
            raise e

    @staticmethod
    def build_context(model_id="test-any-model-id"):
        """Build a gRPC context object containing the specified model ID

        Args:
            model_id(string): The model ID

        Returns:
            context(grpc.ServicerContext): Context object with mm-model-id
        """

        # Create a dummy class for mimicking ServicerContext invocation
        # metadata storage
        class TestContext:
            def __init__(self, model_id):
                self.model_id = model_id
                self.callbacks = []

            def invocation_metadata(self):
                return [("mm-model-id", self.model_id)]

            def add_callback(self, some_function, *args, **kwargs):
                self.callbacks.append(
                    {"func": some_function, "args": args, "kwargs": kwargs}
                )
                return True

            def cancel(self):
                [f["func"](*f["args"], **f["kwargs"]) for f in self.callbacks]

        return TestContext(model_id)

    @staticmethod
    def get_good_model_path():
        return os.path.join(os.path.dirname(__file__), "models", "foo")

    @staticmethod
    def get_other_good_model_path():
        return os.path.join(os.path.dirname(__file__), "models", "bar")

    @staticmethod
    def get_good_model_archive_path():
        return os.path.join(os.path.dirname(__file__), "models", "foo_archive.zip")

    @staticmethod
    def get_bad_model_archive_path():
        return os.path.join(os.path.dirname(__file__), "models", "bad_model.zip")

    @staticmethod
    def get_invalid_model_archive_path():
        return os.path.join(os.path.dirname(__file__), "models", "bad_archive.zip")

    @staticmethod
    def get_good_model_type():
        return "fake_module"



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Scheme for randomly picking 1 augmentor from a list per object.
"""
# Standard
import random

# First Party
import alog

# Local
from ...exceptions import error_handler
from .base import SchemeBase

log = alog.use_channel("RSING_AUG_SCHEME")
error = error_handler.get(log)


class RandomSingleSelectionScheme(SchemeBase):
    def __init__(self, selection_probs, augmentors, random_seed=1001):
        """Create a merging augmentor scheme which randomly applies one of many encapsulated
        augmentors when executed.

        Args:
            selection_probs (list(int|float) | tuple(int|float)): Probability
                values for applying each augmentor (must sum to 1).
            augmentors (list(AugmentorBase) | tuple(AugmentorBase)): Augmentors
                to be applied (in same order as selection_probs).
            random_seed (int): Random seed for controlling shuffling behavior.
        """
        super().__init__(True, augmentors, random_seed)
        error.type_check("<COR26721310E>", list, tuple, selection_probs=selection_probs)
        error.type_check_all(
            "<COR89551931E>", int, float, selection_probs=selection_probs
        )
        error.value_check(
            "<COR22821754E>",
            len(selection_probs) == len(augmentors),
            "Number of selection probabilties must match the number of augmentors",
        )
        error.value_check(
            "<COR82891072E>",
            all(0 <= prob <= 1 for prob in selection_probs),
            "Selection probabilities must be in the range [0, 1]",
        )
        error.value_check(
            "<COR00872610E>",
            sum(selection_probs) == 1,
            "Selection probabilities must sum up to one to create a single selection scheme",
        )
        self._selection_probs = selection_probs

    def _execute(self, obj):
        """Execute the merged scheme by picking one random augmentor and applying it.

        Args:
            obj (str | caikit.core.data_model.DataBase): Object to be augmented.
        Returns:
            str | caikit.core.data_model.DataBase: Augmented object of same type
                as input obj.
        """
        aug = random.choices(self._augmentors, weights=self._selection_probs)[0]
        return aug.augment(obj)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares parent class for more specific test classes.
"""

# Standard
from typing import Any
import os
import threading
import unittest

# Local
from caikit.core.toolkit import logging


class TestCaseBase(unittest.TestCase):
    """Parent class for all specific test classes in Caikit Core."""

    # common reference to the place where we put our fixtures, because a ton of tests use this
    fixtures_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "fixtures")
    # Other things that we use often
    modules_fixtures_dir = os.path.join(fixtures_dir, "modules")
    toolkit_fixtures_dir = os.path.join(fixtures_dir, "toolkit")

    def __init__(self, *args, **kwargs):
        # initialize parent class
        super().__init__(*args, **kwargs)

    def _compare_seqs(self, obj_seq1, obj_seq2, properties):
        """Given two sequences, ensure that they are the same length, and that each specified
        property is equal per object. This method explodes if the sequences are not the same
        as defined by assertEqual on the properties of interest.

        Args:
            obj_seq1: list | tuple
                List or tuple sequence.
            obj_seq2: list | tuple
                List or tuple sequence.
            properties: list(str) | tuple(str)
                attributes whose values will be compared across sequences.
        """
        self.assertEqual(len(obj_seq1), len(obj_seq2))
        for obj1, obj2 in zip(obj_seq1, obj_seq2):
            for prop in properties:
                self.assertEqual(getattr(obj1, prop), getattr(obj2, prop))

    def assert_equals_file(self, obj: Any, file_path: str):
        """Compare the string representation

        Args:
            obj: Any
                Object to compare against the contents of the file
            file_path: str
                Location of the file containing the expected string value of `obj`
        """
        with open(file_path, "r", encoding="utf-8") as file_handle:
            file_contents = file_handle.read()
        self.assertEqual(str(obj), file_contents)


def skip_in_wheel_test(cls):
    """In some cases (scripts, model evaluation), the code may not be shipped as part of the
    compiled wheel. In these situations, we only run the tests in unit tests, not in wheel tests.
    We do this by decorating the class - if env var NO_SOURCE is set to 1, as it is in the wheel
    test env, we return None so that the class is skipped by PyTest. Otherwise we return the class
    so that it can run normally.

    Args:
        cls: type
            Test class to be run if NO_SOURCE is not set to 1.
    Returns:
        type | None
            The test class if this is not running in the wheel tests, otherwise None.
    """
    if os.getenv("NO_SOURCE") is not None and int(os.getenv("NO_SOURCE")) == 1:
        return None
    return cls


class catch_threading_exception:
    """
    catch_threading_exception is only available in Python 3.10+
    Maintaining this copy to run tests with Python 3.8 and 3.9

    From https://github.com/python/cpython/blob/fbc9d0dbb22549bac2706f61f3ab631239d357b4/Lib/test/support/threading_helper.py#LL154C1-L208C24

    Context manager catching threading.Thread exception using
    threading.excepthook.

    Attributes set when an exception is caught:

    * exc_type
    * exc_value
    * exc_traceback
    * thread

    See threading.excepthook() documentation for these attributes.

    These attributes are deleted at the context manager exit.

    Usage:

        with threading_helper.catch_threading_exception() as cm:
            # code spawning a thread which raises an exception
            ...

            # check the thread exception, use cm attributes:
            # exc_type, exc_value, exc_traceback, thread
            ...

        # exc_type, exc_value, exc_traceback, thread attributes of cm no longer
        # exists at this point
        # (to avoid reference cycles)
    """

    def __init__(self):
        self.exc_type = None
        self.exc_value = None
        self.exc_traceback = None
        self.thread = None
        self._old_hook = None

    def _hook(self, args):
        self.exc_type = args.exc_type
        self.exc_value = args.exc_value
        self.exc_traceback = args.exc_traceback
        self.thread = args.thread

    def __enter__(self):
        self._old_hook = threading.excepthook
        threading.excepthook = self._hook
        return self

    def __exit__(self, *exc_info):
        threading.excepthook = self._old_hook
        del self.exc_type
        del self.exc_value
        del self.exc_traceback
        del self.thread



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""This module contains an implementation of a metaclass that hijacks all
${class}.load(...) invocations.

For a good read on metaclasses, see https://realpython.com/python-metaclasses/

Our goal is to have every instance of a `caikit.core.ModuleBase` automatically populated with
metadata when it is constructed, without requiring anything of the module authors.
The source of this metadata will be the `config.yml` file that resides within the module directory
to be loaded. Since we require the path to that file in order to read the model's metadata, we
cannot simply define a base constructor. Instead, we patch over the module's .load() function,
which is guaranteed to be called with a path to a serialized module.

Additionally, the naive solution of manually patching .load functions does not work when
inheritance is involved. For example::

    import caikit.core

    class ParentModule(caikit.core.ModuleBase):

        @classmethod
        def load(cls, module_path):
            return cls()

    class ChildModule(ParentModule):

        @classmethod
        def load(cls, module_path):
            return super().load()

    # This is fine!
    assert isinstance(ChildModule.load(), ChildModule)

    def injector(load_fn):
        def injected_load(*args):
            module = load_fn(*args)
            module.metadata = {"stuff"}
            return module
        return classmethod(injected_load)

    for clz in (ParentModule, ChildModule):
        # But this line binds the new load function directly to each class :(
        clz.load = injector(clz.load)

    # And this will now raise since ParentModule is returned
    assert isinstance(ChildModule.load(), ChildModule)


Instead of binding new metadata-injecting load functions directly to a class at import time, we
need to bind the new load function at contruction time, when the class hierarchy with inheritance
is known.
"""

# Standard
import abc
import functools

# First Party
import alog

# Local
from ..exceptions import error_handler
from .config import ModuleConfig

log = alog.use_channel("METADATA_INJECT")
error = error_handler.get(log)


class _ModuleBaseMeta(abc.ABCMeta):
    """This is the metaclass used by `caikit.core.ModuleBase`.

    This metaclass populates the `metadata` property of any module that is created by invoking
    a `load` classmethod on a derived class.
    """

    # pylint: disable=arguments-differ
    def __new__(mcs, name, bases, attrs):
        real_load = attrs.get("load")
        if real_load is not None:
            log.debug3("Wrapping a load function on class %s", mcs)

            @alog.logged_function(log.trace)
            def metadata_injecting_load(clz, *args, **kwargs):
                """This function is the replacement for the module's original `.load`"""
                path = None
                module_config = None

                # Many load functions rename what the `path` argument is.
                # Usually, it's just the first positional argument.
                # But, `load` may be called with kwargs-only, like:
                # MyModelClass.load(model_path="/some/path")
                log.debug3(
                    "Attempting to find model path from load args: [%s] [%s]",
                    args,
                    kwargs,
                )
                if len(args) > 0:
                    log.debug3(
                        "Using first positional argument to load as model path: %s",
                        args[0],
                    )
                    path = args[0]
                else:
                    for kw, arg in kwargs.items():
                        if "path" in kw:
                            log.debug3(
                                "Using named keyword argument to load as model path: [%s: %s]",
                                kw,
                                arg,
                            )
                            path = arg
                            break
                try:
                    if path:
                        module_config = ModuleConfig.load(path)
                except FileNotFoundError:
                    log.error(
                        "Could not load module metadata while loading with %s %s",
                        args,
                        kwargs,
                    )

                # Call the original .load function to load the module
                module = real_load.__func__(clz, *args, **kwargs)

                # defer any "is this really a module" logic until after the load call
                if hasattr(module, "metadata") and module_config:
                    module.metadata.update(module_config)

                return module

            # Wrap the load function so that the final method appears the same
            # as the original
            metadata_injecting_load = functools.wraps(real_load.__func__)(
                metadata_injecting_load
            )
            attrs["load"] = classmethod(metadata_injecting_load)

        return super().__new__(mcs, name, bases, attrs)

    def __setattr__(cls, name, val):
        """Overwrite __setattr__ to warn on any dynamic updates to the load function.
        We'd rather not lose all the work we did to wrap `.load` with metadata injection
        in the constructor!
        """
        if name == "load":
            # NB: warn instead of throw because some libraries will mock out .load during
            # unit testing where it's easier than trying to build a quick-loading dummy model
            log.warning("Overwriting load on a module will break metadata persistence!")
        return super().__setattr__(name, val)



"""
A hypothetical inner module that transforms some output type rather than taking domain input
"""
# Local
from ...data_model.sample import SampleOutputType
from caikit.core import ModuleSaver
import caikit.core


@caikit.core.module("00110203-baad-beef-0809-0a0b02dd0e0f", "InnerModule", "0.0.1")
class InnerModule(caikit.core.ModuleBase):
    def __init__(self, batch_size=64, learning_rate=0.0015):
        super().__init__()
        self.batch_size = batch_size
        self.learning_rate = learning_rate

    def run(self, some_input: SampleOutputType) -> SampleOutputType:
        return SampleOutputType(f"nested greeting: {some_input.greeting}")

    def save(self, model_path):
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with module_saver:
            pass

    @classmethod
    def load(cls, model_path):
        return cls()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
import json
import os
import traceback

# Third Party
from google.protobuf.json_format import ParseDict
from google.protobuf.message import Message as ProtoMessageType
import grpc

# First Party
import alog

# Local
from caikit import get_config
from caikit.interfaces.common.data_model.stream_sources import (
    Directory,
    File,
    ListOfFiles,
)
from caikit.runtime.protobufs import process_pb2, process_pb2_grpc
from caikit.runtime.service_factory import ServicePackage
from caikit.runtime.service_generation.data_stream_source import DataStreamSourceBase
from caikit.runtime.service_generation.rpcs import ModuleClassTrainRPC
from caikit.runtime.servicers.global_train_servicer import GlobalTrainServicer
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
import caikit.core

log = alog.use_channel("MT-SERVICR-I")


class ModelTrainServicerImpl(process_pb2_grpc.ProcessServicer):
    """This class contains the implementation of the Model Train process.proto
    interface (a.k.a. the Process Servicer).
    """

    def __init__(self, training_service: ServicePackage):
        self._training_service = training_service
        self._gts = GlobalTrainServicer(self._training_service)

    def Run(self, request, context):
        """`Run` RPC -- launches a training job.
        Args:
            request(process_pb2.ProcessRequest):
                Generated from process.proto
            context(grpc.ServicerContext): Context object (contains request metadata, etc)
        Returns:
            process_pb2.ProcessResponse:
                Generated from process.proto
        """
        log.info("<RUN02584562I>", "Calling ModelTrainServicer.Run!")
        log.debug("<RUN91039903D>", "ProcessRequest: %s", request)

        try:
            request_dict = request.request_dict

            # get the model to train
            train_module = caikit.core.registries.module_registry().get(
                request_dict["train_module"]
            )
            log.debug("<RUN76043064D>", "train_module: %s", train_module)
            if train_module is None:
                raise CaikitRuntimeException(
                    grpc.StatusCode.INVALID_ARGUMENT,
                    "Model Train not able to parse module for this Train Request",
                )

            # prepare the model's train request
            training_params = json.loads(request_dict["training_params"])
            request_name = ModuleClassTrainRPC.module_class_to_req_name(train_module)
            log.debug("<RUN22972949D>", "request_name: %s", request_name)

            if not hasattr(self._training_service.messages, request_name):
                raise CaikitRuntimeException(
                    grpc.StatusCode.INTERNAL,
                    f"Model Train not able to create a request for {request_name}",
                )
            request_class = getattr(self._training_service.messages, request_name)
            # construct the protobufs message request with train params
            train_message_request = ParseDict(
                training_params,
                request_class(),
            )
            # update any file references relative to the training input dir
            self._update_file_references(
                train_message_request.parameters, request.training_input_dir
            )

            # Figure out the right training output dir to use for this request.
            # Model Train does not allow this value in the request to be set by
            # the inbound request to the frontend of MT, so the
            # request.training_output_dir value will always point to a private
            # emptyDir mount. In the case where the user wants to save to some
            # other path (e.g. a mounted PVC), the user must set that in the
            # global config as runtime.training.output_dir. This will cause all
            # requests to be saved there and therefore S3 uploads will not be
            # done.
            output_dir = (
                get_config().runtime.training.output_dir or request.training_output_dir
            )

            # make the train call
            log.debug("Training output dir: %s", output_dir)
            training_response = self._gts.run_training_job(
                request=train_message_request,
                module=train_module,
                training_output_dir=output_dir,
                context=context,
                wait=True,
                # TODO: This usage of the server will sit behind Model Train and
                #   will therefore always use a local trainer which supports the
                #   external_training_id override. If that is ever not the case
                #   this will be brokenly passing that kwarg through to the
                #   module's train function.
                external_training_id=request.customTrainingID or request.trainingID,
            )
            log.debug("<RUN00837184D>", "training_response: %s", training_response)
            # return response
            process_response = process_pb2.ProcessResponse(
                trainingID=request.trainingID,
                customTrainingID=request.customTrainingID,
            )
            return process_response

        except MemoryError as e:
            raise CaikitRuntimeException(
                grpc.StatusCode.INTERNAL, "OOM error during training"
            ) from e

        except CaikitRuntimeException as e:
            # pylint: disable=R0801
            log.warning(
                {
                    "log_code": "<RUN50531303W>",
                    "message": e.message,
                    "error_id": e.id,
                    **e.metadata,
                }
            )
            raise e

        except (TypeError, ValueError) as e:
            log.warning(
                {
                    "log_code": "<RUN490449039W>",
                    "message": repr(e),
                    "stack_trace": traceback.format_exc(),
                }
            )
            raise CaikitRuntimeException(
                grpc.StatusCode.INVALID_ARGUMENT,
                f"Exception raised during training. This may be a problem with your input: {e}",
            ) from e

        except Exception as e:
            log.warning(
                {
                    "log_code": "<RUN48349070W>",
                    "message": repr(e),
                    "stack_trace": traceback.format_exc(),
                }
            )
            raise CaikitRuntimeException(
                grpc.StatusCode.INTERNAL, "Unhandled exception during training"
            ) from e

    def _update_file_references(
        self, train_message_request: ProtoMessageType, training_input_dir: str
    ):
        # Check through request and redirect file references to the training input dir
        # This requires a bit of fancy footwork:
        # 1. Get the real data model for the request
        dm_class = caikit.core.data_model.DataBase.get_class_for_proto(
            train_message_request
        )
        training_request_data_model = dm_class.from_proto(train_message_request)
        # 2. Find any data streams
        for attr_name in training_request_data_model.__annotations__.keys():
            val = getattr(training_request_data_model, attr_name)
            if isinstance(val, DataStreamSourceBase):
                # 3. Look for file pointers and update them
                source = val.data_stream
                if isinstance(source, File):
                    source.filename = self._update_file_ref(
                        source.filename, training_input_dir
                    )
                if isinstance(source, ListOfFiles):
                    source.files = [
                        self._update_file_ref(filename, training_input_dir)
                        for filename in source.files
                    ]
                if isinstance(source, Directory):
                    source.dirname = self._update_file_ref(
                        source.dirname, training_input_dir
                    )
                # Importantly we must set the attribute directly back on the proto message
                # Re-serializing the entire request wouldn't properly propagate the existence
                # of unset fields
                proto_val = getattr(train_message_request, attr_name)
                proto_val.CopyFrom(val.to_proto())

    @staticmethod
    def _update_file_ref(file_path: str, training_input_dir: str) -> str:
        """Append the training_input_dir to the file_path if the resulting path exists"""
        updated_path = os.path.join(training_input_dir, file_path)
        if os.path.exists(updated_path):
            return updated_path
        return file_path



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Global factories for model management
"""
# Standard
from typing import Optional
import importlib

# First Party
import alog

# Local
from ..exceptions import error_handler
from ..toolkit.factory import Factory, FactoryConstructible
from .local_model_finder import LocalModelFinder
from .local_model_initializer import LocalModelInitializer
from .local_model_trainer import LocalModelTrainer
from .multi_model_finder import MultiModelFinder

log = alog.use_channel("MMFCTRY")
error = error_handler.get(log)


class ImportableFactory(Factory):
    """An ImportableFactory extends the base Factory to allow the construction
    to specify an "import_class" field that will be used to import and register
    the implementation class before attempting to initialize it.
    """

    IMPORT_CLASS_KEY = "import_class"

    def construct(
        self,
        instance_config: dict,
        instance_name: Optional[str] = None,
    ):
        # Look for an import_class and import and register it if found
        import_class_val = instance_config.get(self.__class__.IMPORT_CLASS_KEY)
        if import_class_val:
            error.type_check(
                "<COR85108801E>",
                str,
                **{self.__class__.IMPORT_CLASS_KEY: import_class_val},
            )
            module_name, class_name = import_class_val.rsplit(".", 1)
            try:
                imported_module = importlib.import_module(module_name)
            except ImportError:
                error(
                    "<COR46837141E>",
                    ValueError(
                        "Invalid {}: Module cannot be imported [{}]".format(
                            self.__class__.IMPORT_CLASS_KEY,
                            module_name,
                        )
                    ),
                )
            try:
                imported_class = getattr(imported_module, class_name)
            except AttributeError:
                error(
                    "<COR46837142E>",
                    ValueError(
                        "Invalid {}: No such class [{}] on module [{}]".format(
                            self.__class__.IMPORT_CLASS_KEY,
                            class_name,
                            module_name,
                        )
                    ),
                )
            error.subclass_check("<COR52306423E>", imported_class, FactoryConstructible)

            self.register(imported_class)
        return super().construct(instance_config, instance_name)


# Model trainer factory. A trainer is responsible for performing the train
# operation against a configured framework connection.
model_trainer_factory = ImportableFactory("ModelTrainer")
model_trainer_factory.register(LocalModelTrainer)

# Model finder factory. A finder is responsible for locating a well defined
# configuration for a model based on a unique path or id.
model_finder_factory = ImportableFactory("ModelFinder")
model_finder_factory.register(LocalModelFinder)
model_finder_factory.register(MultiModelFinder)

# Model initializer factory. An initializer is responsible for taking a model
# configuration and preparing the model to be run in a configured runtime
# location.
model_initializer_factory = ImportableFactory("ModelInitializer")
model_initializer_factory.register(LocalModelInitializer)



"""
A sample module for sample things!
"""
# Standard
from typing import Iterable, List, Optional, Union
import os
import time

# Local
from ...data_model.sample import (
    SampleInputType,
    SampleOutputType,
    SampleTask,
    SampleTrainingType,
)
from caikit.core.data_model import DataStream
from caikit.core.modules import ModuleLoader, ModuleSaver
import caikit.core


@caikit.core.module(
    "00110203-0405-0607-0809-0a0b02dd0e0f", "SampleModule", "0.0.1", SampleTask
)
class SampleModule(caikit.core.ModuleBase):
    POISON_PILL_NAME = "Bob Marley"
    POISON_PILL_BATCH_SIZE = 999

    def __init__(self, batch_size=64, learning_rate=0.0015, stream_size=10):
        super().__init__()
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.stream_size = stream_size

    @classmethod
    def load(cls, model_path, **kwargs):
        loader = ModuleLoader(model_path)
        config = loader.config
        return cls(config["train"]["batch_size"], config["train"]["learning_rate"])

    @SampleTask.taskmethod()
    def run(
        self, sample_input: SampleInputType, throw: bool = False
    ) -> SampleOutputType:
        """
        Args:
            sample_input (sample_lib.data_model.SampleInputType): the input

        Returns:
            sample_lib.data_model.SampleOutputType: The output
        """
        if throw:
            raise RuntimeError("barf!")
        assert isinstance(sample_input, SampleInputType)
        if sample_input.name == self.POISON_PILL_NAME:
            raise ValueError(f"{self.POISON_PILL_NAME} is not allowed!")
        return SampleOutputType(f"Hello {sample_input.name}")

    @SampleTask.taskmethod(output_streaming=True)
    def run_stream_out(
        self, sample_input: SampleInputType
    ) -> DataStream[SampleOutputType]:
        """
        Args:
            sample_input (sample_lib.data_model.SampleInputType): the input

        Returns:
            caikit.core.data_model.DataStream[sample_lib.data_model.SampleOutputType]: The output
                stream
        """
        list_ = [
            SampleOutputType(f"Hello {sample_input.name} stream")
            for x in range(self.stream_size)
        ]
        stream = DataStream.from_iterable(list_)
        return stream

    @SampleTask.taskmethod(input_streaming=True, output_streaming=True)
    def run_bidi_stream(
        self, sample_inputs: DataStream[SampleInputType]
    ) -> Iterable[SampleOutputType]:
        """
        Args:
            sample_inputs (caikit.core.data_model.DataStream[sample_lib.data_model.SampleInputType]): the input

        Returns:
            caikit.core.data_model.DataStream[sample_lib.data_model.SampleOutputType]: The output
                stream
        """
        for sample_input in sample_inputs:
            yield self.run(sample_input)

    def save(self, model_path, **kwargs):
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with module_saver:
            config_options = {
                "train": {
                    "batch_size": self.batch_size,
                    "learning_rate": self.learning_rate,
                },
            }
            if "keyword" in kwargs:
                config_options["test_keyword"] = True

            module_saver.update_config(config_options)

    @classmethod
    def train(
        cls,
        training_data: DataStream[SampleTrainingType],
        union_list: Optional[Union[List[str], List[int]]] = None,
        batch_size: int = 64,
        oom_exit: bool = False,
        oom_exit_code: Optional[int] = None,
        sleep_time: float = 0,
        sleep_increment: float = 0.001,
        **kwargs,
    ) -> "SampleModule":
        """Sample training method that produces a trained model"""

        # If requested, set an event to indicate that the training has started
        start_event = kwargs.get("start_event")
        if start_event is not None:
            start_event.set()

        # If needed, wait for an event
        # NOTE: We need to pull this from **kwargs because it's not a valid arg
        #   for train API deduction. It's only needed for testing purposes, so
        #   this is definitely a non-standard usage pattern!
        wait_event = kwargs.get("wait_event")
        if wait_event is not None:
            # Set a timeout here so tests don't hang forever in failure states
            wait_event.wait(timeout=1)

        # If needed, wait for a long time
        # NOTE: DestroyableThread is a "best effort" at destroying a threaded
        #   work and is not actually capable of destroying many type of work. If
        #   this is written as time.sleep(sleep_time), the interrupt will not
        #   work.
        if sleep_time:
            start_time = time.time()
            while time.time() - sleep_time < start_time:
                time.sleep(sleep_increment)

        if oom_exit:
            # exit with OOM code. Note _exit method is used to exit the
            # process with specified status without calling cleanup handlers
            # to replicate OOM scenario
            if oom_exit_code:
                os._exit(oom_exit_code)
            else:
                os._exit(137)

        if batch_size == cls.POISON_PILL_BATCH_SIZE:
            raise ValueError(
                f"Batch size of {cls.POISON_PILL_BATCH_SIZE} is not allowed!"
            )
        # Barf if we were incorrectly passed data not in datastream format
        assert isinstance(training_data, DataStream)
        for _ in training_data:
            # Consume the stream
            pass
        if union_list:
            assert isinstance(union_list, List)
            assert len(union_list) > 0
        return cls(batch_size=batch_size)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General global train servicer tests - run for both subprocess and local"""
# Standard
from tempfile import TemporaryDirectory
from unittest.mock import patch
import multiprocessing
import os
import threading
import time

# Third Party
import grpc
import pytest

# Local
from caikit.config import get_config
from caikit.core import MODEL_MANAGER
from caikit.core.data_model.producer import ProducerId
from caikit.interfaces.common.data_model.stream_sources import S3Path
from caikit.runtime.service_factory import (
    get_inference_request,
    get_train_params,
    get_train_request,
)
from caikit.runtime.servicers.global_train_servicer import GlobalTrainServicer
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from sample_lib.data_model.sample import (
    OtherOutputType,
    SampleInputType,
    SampleOutputType,
    SampleTrainingType,
)
from sample_lib.modules import CompositeModule, OtherModule, SampleModule
from tests.conftest import random_test_id, set_use_subprocess
from tests.fixtures import Fixtures
from tests.runtime.conftest import register_trained_model
import caikit.core

## Helpers #####################################################################

HAPPY_PATH_INPUT_DM = SampleInputType(name="Gabe")


@pytest.fixture(autouse=True, params=[True, False])
def set_train_location(request, sample_train_servicer, reset_model_manager):
    """This fixture ensures that all tests in this file will be run with both
    subprocess and local training styles
    """
    with set_use_subprocess(request.param):
        yield


# Train tests for the GlobalTrainServicer class ############################################################
##############
# Normal cases
##############


def test_global_train_sample_task(
    sample_train_service,
    sample_train_servicer,
    sample_inference_service,
    sample_predict_servicer,
    sample_task_unary_rpc,
):
    """Global train of TrainRequest returns a training job with the correct
    model name, and some training id for a basic train function that doesn't
    require any loaded model
    """
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(data=[SampleTrainingType(1)])
    )
    model_name = random_test_id()
    train_class = get_train_request(SampleModule)
    train_request_params_class = get_train_params(SampleModule)
    train_request = train_class(
        model_name=model_name,
        parameters=train_request_params_class(
            batch_size=42,
            training_data=training_data,
        ),
    ).to_proto()

    training_response = sample_train_servicer.Train(
        train_request, Fixtures.build_context("foo")
    )

    assert training_response.model_name == model_name

    assert training_response.training_id is not None
    assert isinstance(training_response.training_id, str)
    register_trained_model(
        sample_predict_servicer,
        training_response.model_name,
        training_response.training_id,
    )

    model_future = MODEL_MANAGER.get_model_future(training_response.training_id)
    expected_save_path_bits = os.path.join(
        training_response.training_id, training_response.model_name
    )
    assert expected_save_path_bits in model_future.save_path

    result = model_future.load()
    assert result.batch_size == 42
    assert (
        result.MODULE_CLASS
        == "sample_lib.modules.sample_task.sample_implementation.SampleModule"
    )

    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    inference_response = sample_predict_servicer.Predict(
        predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
        Fixtures.build_context(training_response.model_name),
        caikit_rpc=sample_task_unary_rpc,
    )
    assert (
        inference_response
        == SampleOutputType(
            greeting="Hello Gabe",
        ).to_proto()
    )


def test_global_train_other_task(
    sample_train_service,
    sample_train_servicer,
    sample_inference_service,
    sample_predict_servicer,
):
    """Global train of TrainRequest returns a training job with the correct
    model name, and some training id for a basic train function that doesn't
    require any loaded model
    """
    batch_size = 42
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceInt
    training_data = stream_type(jsondata=stream_type.JsonData(data=[1]))
    train_class = get_train_request(OtherModule)
    train_request_params_class = get_train_params(OtherModule)
    train_request = train_class(
        model_name="Other module Training",
        parameters=train_request_params_class(
            training_data=training_data,
            sample_input_sampleinputtype=HAPPY_PATH_INPUT_DM,
            batch_size=batch_size,
        ),
    ).to_proto()

    training_response = sample_train_servicer.Train(
        train_request, Fixtures.build_context("foo")
    )
    assert training_response.model_name == "Other module Training"
    assert training_response.training_id is not None
    assert isinstance(training_response.training_id, str)
    register_trained_model(
        sample_predict_servicer,
        training_response.model_name,
        training_response.training_id,
    )

    result = MODEL_MANAGER.get_model_future(training_response.training_id).load()
    assert result.batch_size == batch_size
    assert (
        result.MODULE_CLASS
        == "sample_lib.modules.other_task.other_implementation.OtherModule"
    )

    predict_class = get_inference_request(OtherModule.TASK_CLASS)
    inference_response = sample_predict_servicer.Predict(
        predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
        Fixtures.build_context(training_response.model_name),
        caikit_rpc=sample_inference_service.caikit_rpcs["OtherTaskPredict"],
    )
    assert (
        inference_response
        == OtherOutputType(
            farewell=f"goodbye: Gabe {batch_size} times",
            producer_id=ProducerId("other_mod", "1.1.1"),
        ).to_proto()
    )


def test_global_train_Another_Widget_that_requires_SampleWidget_loaded_should_not_raise(
    sample_task_model_id,
    sample_train_service,
    sample_train_servicer,
    sample_inference_service,
    sample_predict_servicer,
    sample_task_unary_rpc,
):
    """Global train of TrainRequest returns a training job with the correct model name, and some training id for a train function that requires another loaded model"""
    sample_model = caikit.interfaces.runtime.data_model.ModelPointer(
        model_id=sample_task_model_id
    )

    train_class = get_train_request(CompositeModule)
    train_request_params_class = get_train_params(CompositeModule)
    training_request = train_class(
        model_name="AnotherWidget_Training",
        parameters=train_request_params_class(sample_block=sample_model),
    ).to_proto()

    training_response = sample_train_servicer.Train(
        training_request, Fixtures.build_context("foo")
    )
    assert training_response.model_name == "AnotherWidget_Training"
    assert training_response.training_id is not None
    assert isinstance(training_response.training_id, str)
    register_trained_model(
        sample_predict_servicer,
        training_response.model_name,
        training_response.training_id,
    )

    training_result = MODEL_MANAGER.get_model_future(
        training_response.training_id
    ).load()
    assert (
        training_result.MODULE_CLASS
        == "sample_lib.modules.sample_task.composite_module.CompositeModule"
    )

    # make sure the trained model can run inference
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    inference_response = sample_predict_servicer.Predict(
        predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
        Fixtures.build_context(training_response.model_name),
        caikit_rpc=sample_task_unary_rpc,
    )
    assert (
        inference_response
        == SampleOutputType(
            greeting="Hello Gabe",
        ).to_proto()
    )


def test_run_train_job_works_with_wait(
    sample_train_service,
    sample_inference_service,
    sample_predict_servicer,
    sample_task_unary_rpc,
):
    """Check if run_train_job works as expected for syncronous requests"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(data=[SampleTrainingType(1)])
    )
    train_class = get_train_request(SampleModule)
    train_request_params_class = get_train_params(SampleModule)
    train_request = train_class(
        model_name=random_test_id(),
        parameters=train_request_params_class(
            batch_size=42,
            training_data=training_data,
        ),
    ).to_proto()
    servicer = GlobalTrainServicer(training_service=sample_train_service)
    with TemporaryDirectory() as tmp_dir:
        training_response = servicer.run_training_job(
            train_request,
            SampleModule,
            training_output_dir=tmp_dir,
            context=Fixtures.build_context("foo"),
            wait=True,
        )
        register_trained_model(
            sample_predict_servicer,
            training_response.model_name,
            training_response.training_id,
        )

        predict_class = get_inference_request(SampleModule.TASK_CLASS)
        inference_response = sample_predict_servicer.Predict(
            predict_class(sample_input=SampleInputType(name="Test")).to_proto(),
            Fixtures.build_context(training_response.model_name),
            caikit_rpc=sample_task_unary_rpc,
        )
        assert (
            inference_response
            == SampleOutputType(
                greeting="Hello Test",
            ).to_proto()
        )


#############
# Error cases
#############


def test_global_train_Another_Widget_that_requires_SampleWidget_but_not_loaded_should_raise(
    sample_train_service, sample_train_servicer
):
    """Global train of TrainRequest raises when calling a train function that requires another loaded model, but model is not loaded"""
    model_id = random_test_id()

    sample_model = caikit.interfaces.runtime.data_model.ModelPointer(model_id=model_id)
    train_class = get_train_request(CompositeModule)
    train_request_params_class = get_train_params(CompositeModule)
    request = train_class(
        model_name="AnotherWidget_Training",
        parameters=train_request_params_class(sample_block=sample_model),
    ).to_proto()

    with pytest.raises(CaikitRuntimeException) as context:
        sample_train_servicer.Train(request, Fixtures.build_context("foo"))

    assert f"Model '{model_id}' not loaded" == context.value.message


def test_global_train_Edge_Case_Widget_should_raise_when_error_surfaces_from_module(
    sample_train_service, sample_train_servicer
):
    """Test that if a module raises a ValueError, we should surface it to the user in a helpful way"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(data=[SampleTrainingType(1)])
    )

    train_class = get_train_request(SampleModule)
    train_request_params_class = get_train_params(SampleModule)
    train_request = train_class(
        model_name=random_test_id(),
        parameters=train_request_params_class(
            batch_size=999,
            training_data=training_data,
        ),
    ).to_proto()

    training_response = sample_train_servicer.Train(
        train_request, Fixtures.build_context("foo")
    )
    MODEL_MANAGER.get_model_future(training_response.training_id).wait()

    future_info = MODEL_MANAGER.get_model_future(
        training_response.training_id
    ).get_info()
    assert f"Batch size of 999 is not allowed!" in str(future_info.errors[0])


def test_local_trainer_rejects_s3_output_paths(
    sample_train_service, sample_train_servicer
):
    """Test that if an output_path references S3 but the trainer does not know how to handle it, an error is raised"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(data=[SampleTrainingType(1)])
    )
    train_class = get_train_request(SampleModule)
    train_request_params_class = get_train_params(SampleModule)
    train_request = train_class(
        model_name=random_test_id(),
        output_path=S3Path(path="foo"),
        parameters=train_request_params_class(
            batch_size=42,
            training_data=training_data,
            oom_exit=True,
        ),
    ).to_proto()

    with pytest.raises(
        CaikitRuntimeException, match=".*S3 output path not supported by this runtime"
    ) as ctx:
        sample_train_servicer.Train(train_request, Fixtures.build_context("foo"))
    assert ctx.value.status_code == grpc.StatusCode.INVALID_ARGUMENT


#####################################################################


@pytest.mark.skip(
    reason="This test fails intermittently. Functionality has to be debugged for race condition"
)
def test_global_train_aborts_long_running_trains(
    sample_train_service, sample_train_servicer
):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(data=[SampleTrainingType(1)])
    )
    training_id = random_test_id()

    train_class = get_train_request(SampleModule)
    train_request_params_class = get_train_params(SampleModule)
    train_request = train_class(
        model_name=training_id,
        parameters=train_request_params_class(
            batch_size=42,
            training_data=training_data,
            oom_exit=False,
        ),
    ).to_proto()

    if sample_train_servicer.use_subprocess:
        start_method = (
            get_config().model_management.trainers.default.config.subprocess_start_method
        )
        test_event = multiprocessing.get_context(start_method).Event()
    else:
        test_event = threading.Event()

    def never_respond(*args, **kwargs):
        """Never ending function"""
        test_event.set()
        while True:
            time.sleep(0.01)

    context = Fixtures.build_context("test-any-unresponsive-model")

    servicer = GlobalTrainServicer(training_service=sample_train_service)

    with TemporaryDirectory() as tmp_dir:
        training_output_dir = tmp_dir

        train_thread = threading.Thread(
            target=servicer.run_training_job,
            args=(
                train_request,
                SampleModule,
                training_id,
                training_output_dir,
                context,
            ),
            kwargs={
                "wait": True,
            },
        )

    # NOTE: We are configuring following timeout
    # to avoid tests from hanging
    request_timeout = 4
    test_event_timeout = 2
    with patch(
        f"{SampleModule.__module__}.{SampleModule.train.__qualname__}",
        never_respond,
    ):
        train_thread.start()
        # NB: assert is here to make sure we called the patched train
        assert test_event.wait(test_event_timeout)

        # Simulate a timeout or client abort
        context.cancel()
        train_thread.join(request_timeout)

    # Make sure the training job actually stopped
    assert not train_thread.is_alive()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import csv
import json
import os
import tempfile
import uuid

# Local
from caikit.core.data_model import DataStream
from caikit.core.data_model.streams.resolver import DataStreamResolver
from caikit.core.exceptions.validation_error import DataValidationError
from tests.base import TestCaseBase


class TestDataStreamResolver(TestCaseBase):
    resolver = DataStreamResolver(list, {"foo": str, "bar": str})

    @staticmethod
    def _randomData():
        lists = [[str(uuid.uuid4()), str(uuid.uuid4())] for _ in range(100)]
        dicts = [{"foo": each_list[0], "bar": each_list[1]} for each_list in lists]
        return lists, dicts

    def test_it_passes_back_a_data_stream(self):
        lists, dicts = self._randomData()
        data_stream = DataStream.from_iterable(lists)
        self.assertEqual(
            list(data_stream), list(self.resolver.as_data_stream(data_stream))
        )

    def test_it_can_convert_streams(self):
        lists, dicts = self._randomData()
        result = self.resolver.as_data_stream(DataStream.from_iterable(dicts))
        self.assertEqual(lists, list(result))

    def test_it_can_handle_json_files(self):
        lists, dicts = self._randomData()
        with tempfile.TemporaryDirectory() as tempdir:
            filename = os.path.join(tempdir, "foo.json")
            with open(filename, "w") as f:
                json.dump(dicts, f)

            self.assertEqual(lists, list(self.resolver.as_data_stream(filename)))

    def test_it_can_handle_csv_files(self):
        lists, dicts = self._randomData()
        with tempfile.TemporaryDirectory() as tempdir:
            filename = os.path.join(tempdir, "foo.csv")
            with open(filename, "w") as f:
                csv.writer(f).writerows(lists)
            self.assertEqual(lists, list(self.resolver.as_data_stream(filename)))

    def test_it_can_handle_csv_formats_with_variable_rows(self):
        lists = [
            ["some text", "label 1"],
            ["some more text", "label 1", "label 2", "label 3"],
        ]
        expected_stream = [
            {"text": "some text", "labels": ["label 1"]},
            {"text": "some more text", "labels": ["label 1", "label 2", "label 3"]},
        ]
        resolver = DataStreamResolver(dict, {"text": str, "labels": list})

        with tempfile.TemporaryDirectory() as tempdir:
            filename = os.path.join(tempdir, "foo.csv")
            with open(filename, "w") as f:
                csv.writer(f).writerows(lists)
            self.assertEqual(expected_stream, list(resolver.as_data_stream(filename)))

    def test_will_throw_data_validation_error_on_bad_data(self):
        sad_resolver = DataStreamResolver(list, {"foo": str, "baz": str})
        lists, dicts = self._randomData()
        result = sad_resolver.as_data_stream(DataStream.from_iterable(dicts))
        with self.assertRaises(DataValidationError):
            list(result)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from caikit.interfaces.nlp.data_model import (
    ClassificationResult,
    ClassificationResults,
    ClassificationTrainRecord,
    ClassifiedGeneratedTextResult,
    FinishReason,
    TokenClassificationResult,
    TokenClassificationResults,
)

## Setup #########################################################################

classification1 = ClassificationResult(label="temperature", score=0.71)

classification2 = ClassificationResult(label="conditions", score=0.98)

classification_result = ClassificationResults(
    results=[classification1, classification2]
)

token_classification1 = TokenClassificationResult(
    start=0, end=5, word="moose", entity="animal", score=0.8
)
token_classification2 = TokenClassificationResult(
    start=7, end=12, word="goose", entity="animal", score=0.7
)
token_classification3 = TokenClassificationResult(
    start=0, end=5, word="llama", entity="animal", score=0.2
)
token_classification_result = TokenClassificationResults(
    results=[token_classification1, token_classification2]
)

classification_train_record = ClassificationTrainRecord(
    text="It is 20 degrees today", labels=["temperature"]
)

classification_generated_text_result = ClassifiedGeneratedTextResult(
    generated_text="moose goose foo bar",
    token_classification_results=ClassifiedGeneratedTextResult.TextGenTokenClassificationResults(
        input=[token_classification3],
        output=[token_classification1, token_classification2],
    ),
    input_token_count=4,
    generated_token_count=7,
    finish_reason=FinishReason.STOP_SEQUENCE,
    seed=42,
)

## Tests ########################################################################

### ClassificationResult


def test_classification_all_fields_accessible():
    classification_result = ClassificationResult(label="temperature", score=0.71)
    assert classification_result.label == "temperature"
    assert classification_result.score == 0.71


def test_classification_from_proto_and_back():
    new = ClassificationResult.from_proto(classification1.to_proto())
    assert new.label == "temperature"
    assert new.score == 0.71


def test_classification_from_json_and_back():
    new = ClassificationResult.from_json(classification1.to_json())
    assert new.label == "temperature"
    assert new.score == 0.71


### ClassificationResults


def test_classification_result_all_fields_accessible():
    classification_result = ClassificationResults(results=[classification1])
    assert classification_result.results[0].label == "temperature"
    assert classification_result.results[0].score == 0.71


def test_classification_result_from_proto_and_back():
    new = ClassificationResults.from_proto(classification_result.to_proto())
    assert new.results[0].label == "temperature"
    assert new.results[0].score == 0.71
    assert new.results[1].label == "conditions"
    assert new.results[1].score == 0.98


def test_classification_result_from_json_and_back():
    new = ClassificationResults.from_json(classification_result.to_json())
    assert new.results[0].label == "temperature"
    assert new.results[0].score == 0.71
    assert new.results[1].label == "conditions"
    assert new.results[1].score == 0.98


### TokenClassificationResult


def test_token_classification_all_fields_accessible():
    token_classification = TokenClassificationResult(
        start=0,
        end=28,
        word="The cow jumped over the moon",
        entity="neutral",
        score=0.6,
    )
    assert token_classification.start == 0
    assert token_classification.end == 28
    assert token_classification.word == "The cow jumped over the moon"
    assert token_classification.entity == "neutral"
    assert token_classification.score == 0.6


def test_classification_from_proto_and_back():
    new = TokenClassificationResult.from_proto(token_classification1.to_proto())
    assert new.start == 0
    assert new.word == "moose"
    assert new.score == 0.8


def test_classification_from_json_and_back():
    new = TokenClassificationResult.from_json(token_classification1.to_json())
    assert new.start == 0
    assert new.word == "moose"
    assert new.score == 0.8


### TokenClassificationResults


def test_token_classification_result_all_fields_accessible():
    token_classification_result = TokenClassificationResults(
        results=[token_classification1]
    )
    assert token_classification_result.results[0].start == 0
    assert token_classification_result.results[0].word == "moose"
    assert token_classification_result.results[0].score == 0.8


def test_token_classification_result_from_proto_and_back():
    new = TokenClassificationResults.from_proto(token_classification_result.to_proto())
    assert new.results[0].start == 0
    assert new.results[0].word == "moose"
    assert new.results[0].score == 0.8
    assert new.results[1].end == 12
    assert new.results[1].entity == "animal"


def test_classification_result_from_json_and_back():
    new = TokenClassificationResults.from_json(token_classification_result.to_json())
    assert new.results[0].start == 0
    assert new.results[0].word == "moose"
    assert new.results[0].score == 0.8
    assert new.results[1].end == 12
    assert new.results[1].entity == "animal"


### ClassificationTrainRecord


def test_classification_train_record_all_fields_accessible():
    classification_train_record = ClassificationTrainRecord(
        text="It is 20 degrees today", labels=["temperature"]
    )
    assert classification_train_record.text == "It is 20 degrees today"
    assert classification_train_record.labels == ["temperature"]


def test_classification_train_record_from_proto_and_back():
    new = ClassificationTrainRecord.from_proto(classification_train_record.to_proto())
    assert new.text == "It is 20 degrees today"
    assert new.labels == ["temperature"]


def test_classification_train_record_from_json_and_back():
    new = ClassificationTrainRecord.from_json(classification_train_record.to_json())
    assert new.text == "It is 20 degrees today"
    assert new.labels == ["temperature"]


### ClassifiedGeneratedTextResult


def test_classification_generated_text_result_all_fields_accessible():
    classification_generated_text_result = ClassifiedGeneratedTextResult(
        generated_text="moose goose foo bar",
        token_classification_results=ClassifiedGeneratedTextResult.TextGenTokenClassificationResults(
            output=[
                token_classification1,
                token_classification2,
            ]
        ),
    )
    assert classification_generated_text_result.generated_text == "moose goose foo bar"
    assert (
        classification_generated_text_result.token_classification_results.output[0]
        == token_classification1
    )
    assert (
        classification_generated_text_result.token_classification_results.output[1]
        == token_classification2
    )


def test_classification_generated_text_result_from_proto_and_back():
    new = ClassifiedGeneratedTextResult.from_proto(
        classification_generated_text_result.to_proto()
    )
    _validate_classification_generated_text_result(new)


def test_classification_generated_text_result_from_json_and_back():
    new = ClassifiedGeneratedTextResult.from_json(
        classification_generated_text_result.to_json()
    )
    _validate_classification_generated_text_result(new)


def _validate_classification_generated_text_result(obj):
    assert obj.generated_text == "moose goose foo bar"

    assert len(obj.token_classification_results.input) == 1
    assert obj.token_classification_results.input[0].start == 0
    assert obj.token_classification_results.input[0].end == 5
    assert obj.token_classification_results.input[0].word == "llama"
    assert obj.token_classification_results.input[0].entity == "animal"
    assert obj.token_classification_results.input[0].score == 0.2

    assert len(obj.token_classification_results.output) == 2
    assert obj.token_classification_results.output[0].start == 0
    assert obj.token_classification_results.output[0].end == 5
    assert obj.token_classification_results.output[0].word == "moose"
    assert obj.token_classification_results.output[0].entity == "animal"
    assert obj.token_classification_results.output[0].score == 0.8

    assert obj.token_classification_results.output[1].start == 7
    assert obj.token_classification_results.output[1].end == 12
    assert obj.token_classification_results.output[1].word == "goose"
    assert obj.token_classification_results.output[1].entity == "animal"
    assert obj.token_classification_results.output[1].score == 0.7

    assert obj.input_token_count == 4
    assert obj.generated_token_count == 7
    assert obj.finish_reason == 5
    assert obj.seed == 42



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from typing import List, Optional, Union

# Local
from caikit.core.data_model import DataStream
from caikit.runtime.service_generation.type_helpers import (
    get_data_stream_type,
    has_data_stream,
    is_model_type,
)
import sample_lib


def test_data_stream_helpers():
    assert has_data_stream(DataStream[int])
    assert has_data_stream(Optional[DataStream[int]])
    assert has_data_stream(Union[List[DataStream[int]], str])

    assert get_data_stream_type(Optional[DataStream[int]]) == DataStream[int]
    assert get_data_stream_type(Union[List[DataStream[str]], int]) == DataStream[str]


def test_model_type_helpers():
    assert is_model_type(sample_lib.modules.sample_task.SampleModule)
    assert is_model_type(Union[str, sample_lib.modules.sample_task.SampleModule])



# Standard
from typing import Dict, List, Optional, Union
import json

# Third Party
import pytest

# First Party
from py_to_proto.dataclass_to_proto import Annotated

# Local
from caikit.core.data_model.base import DataBase
from caikit.runtime.service_generation.protoable import (
    get_protoable_return_type,
    to_protoable_signature,
)
from sample_lib.data_model import SampleInputType, SampleOutputType
import caikit


# Class that does not have to/from_proto
class NonProtoable:
    pass


################################################################
# to_protoable_signature
################################################################


def test_to_primitive_signature_raw():
    assert to_protoable_signature(
        signature={"name": str},
    ) == {"name": str}


def test_to_protoable_signature_union_raw():
    assert to_protoable_signature(
        signature={"name": Union[str, int]},
    ) == {"name": Union[str, int]}


def test_to_protoable_signature_lists():
    assert to_protoable_signature(
        signature={"good_list": List[str], "bad_list": List, "lowercase_l_list": list},
    ) == {"good_list": List[str]}


def test_to_protoable_signature_dm():
    assert to_protoable_signature(
        signature={"name": SampleInputType},
    ) == {"name": SampleInputType}


def test_to_protoable_signature_union_dm():
    assert to_protoable_signature(
        signature={"name": Union[SampleInputType, str]},
    ) == {"name": Union[SampleInputType, str]}


def test_to_protoable_signature_unsupported_type_in_union():
    assert to_protoable_signature(
        signature={"name": Union[NonProtoable, str]},
    ) == {"name": str}


def test_to_protoable_signature_no_dm_types_in_union():
    class AnotherNonProtoable:
        pass

    assert to_protoable_signature(
        signature={"name": Union[NonProtoable, AnotherNonProtoable, str]},
    ) == {"name": str}


def test_to_protoable_signature_multiple_types_in_union():
    """We have the first arg as a supported DM arg, and the last as
    a supported primitive arg. We return the first supported DM arg"""
    assert to_protoable_signature(
        signature={"name": Union[SampleInputType, NonProtoable, str]},
    ) == {"name": SampleInputType}


def test_to_protoable_signature_multiple_no_dm_types_in_union():
    """We have the first arg that's not supported, and 2 primitive args. We
    return the first primitive arg"""
    assert to_protoable_signature(
        signature={"name": Union[NonProtoable, str, int]},
    ) == {"name": str}


def test_to_protoable_signature_no_protoable_types():
    assert (
        to_protoable_signature(
            signature={"name": NonProtoable},
        )
        == {}
    )


def test_to_protoable_signature_dict():
    assert to_protoable_signature(
        signature={"name": Dict[str, int]},
    ) == {"name": Dict[str, int]}


def test_to_protoable_signature_dict_int_keys():
    assert to_protoable_signature(
        signature={"name": Dict[int, float]},
    ) == {"name": Dict[int, float]}


def test_to_protoable_signature_unsupported_dict():
    assert (
        to_protoable_signature(
            signature={"name": Dict[str, NonProtoable]},
        )
        == {}
    )


def test_to_protoable_signature_dict_incomplete_type_hint():
    assert (
        to_protoable_signature(
            signature={"name": dict},
        )
        == {}
    )


def test_to_protoable_signature_optional_list():
    union_list_signature = to_protoable_signature(
        signature={"optional_param": Optional[List[str]]},
    )
    assert union_list_signature == {"optional_param": List[str]}


def test_to_protoable_signature_list_w_primitive():
    union_list_signature = to_protoable_signature(
        signature={"union_list": Union[List[str], int]},
    )
    assert union_list_signature == {
        "union_list": Union[
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.StrSequence,
                "union_list_str_sequence",
            ],
            int,
        ]
    }


def test_to_protoable_signature_multiple_same_type_union_list():
    union_list_signature = to_protoable_signature(
        signature={
            "union_list_arg": Union[List[str], List[int]],
            "union_list_another_arg": Union[List[str], List[int]],
        },
    )
    assert_sig = {
        "union_list_arg": Union[
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.StrSequence,
                "union_list_arg_str_sequence",
            ],
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.IntSequence,
                "union_list_arg_int_sequence",
            ],
        ],
        "union_list_another_arg": Union[
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.StrSequence,
                "union_list_another_arg_str_sequence",
            ],
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.IntSequence,
                "union_list_another_arg_int_sequence",
            ],
        ],
    }
    assert all((union_list_signature.get(k) == v for k, v in assert_sig.items()))


def test_to_protoable_signature_all_types_union_list():
    union_list_signature = to_protoable_signature(
        signature={
            "union_list_arg": Union[List[str], List[int], List[float], List[bool]],
        },
    )
    assert union_list_signature == {
        "union_list_arg": Union[
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.StrSequence,
                "union_list_arg_str_sequence",
            ],
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.IntSequence,
                "union_list_arg_int_sequence",
            ],
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.FloatSequence,
                "union_list_arg_float_sequence",
            ],
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.BoolSequence,
                "union_list_arg_bool_sequence",
            ],
        ]
    }


def test_to_protoable_signature_throws_unknown_type():
    # Third Party
    import numpy as np

    with pytest.raises(AttributeError):
        to_protoable_signature(
            signature={
                "union_list_arg": Union[List[np.int32], List[int]],
            },
        )


def test_to_protoable_signature_empty_list_arg():
    """Empty list arg is ignored by is_protoable_type"""
    assert to_protoable_signature(
        signature={
            "union_list_arg": Union[List, int],
        },
    ) == {"union_list_arg": int}


def test_to_protoable_signature_multiple_diff_type_union_list():
    union_list_signature = to_protoable_signature(
        signature={
            "union_list_arg": Union[List[str], List[int]],
            "union_list_another_arg": Union[List[str], List[bool]],
        },
    )
    assert_sig = {
        "union_list_arg": Union[
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.StrSequence,
                "union_list_arg_str_sequence",
            ],
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.IntSequence,
                "union_list_arg_int_sequence",
            ],
        ],
        "union_list_another_arg": Union[
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.StrSequence,
                "union_list_another_arg_str_sequence",
            ],
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.BoolSequence,
                "union_list_another_arg_bool_sequence",
            ],
        ],
    }
    assert all((union_list_signature.get(k) == v for k, v in assert_sig.items()))


def test_to_protoable_signature_union_list():
    union_list_signature = to_protoable_signature(
        signature={
            "union_list_arg": Union[List[str], List[int]],
        },
    )
    assert union_list_signature == {
        "union_list_arg": Union[
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.StrSequence,
                "union_list_arg_str_sequence",
            ],
            Annotated[
                caikit.interfaces.common.data_model.primitive_sequences.IntSequence,
                "union_list_arg_int_sequence",
            ],
        ]
    }

    # str sequence
    str_seq = caikit.interfaces.common.data_model.primitive_sequences.StrSequence(
        values=["one", "two"]
    )

    ## proto test
    str_seq.from_proto(str_seq.to_proto()) == str_seq
    ## json test
    union_list_str_json_repr = {"values": ["one", "two"]}
    assert str_seq.to_json() == json.dumps(union_list_str_json_repr)
    assert str_seq.from_json(union_list_str_json_repr) == str_seq

    # int Intuence
    int_seq = caikit.interfaces.common.data_model.primitive_sequences.IntSequence(
        values=[1, 2]
    )

    ## proto test
    int_seq.from_proto(int_seq.to_proto()) == int_seq
    ## json test
    union_list_int_json_repr = {"values": [1, 2]}
    assert int_seq.to_json() == json.dumps(union_list_int_json_repr)
    assert int_seq.from_json(union_list_int_json_repr) == int_seq


################################################################
# get_protoable_return_type
################################################################


def test_to_output_dm_type_with_None():
    assert get_protoable_return_type(None) == None


def test_to_output_dm_type_with_raw_primitive():
    assert get_protoable_return_type(str) == str


def test_to_output_dm_type_with_dm():
    assert get_protoable_return_type(SampleOutputType) == SampleOutputType


def test_to_output_dm_type_with_union_dm():
    assert get_protoable_return_type(Union[SampleOutputType, str]) == SampleOutputType


def test_to_output_dm_type_with_union_optional_dm():
    assert (
        get_protoable_return_type(Union[Optional[SampleOutputType], str])
        == SampleOutputType
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from typing import Optional, Type
import os
import uuid

# Local
from caikit.core import MODEL_MANAGER
from caikit.core.data_model import TrainingStatus
from caikit.core.model_management import (
    ModelFinderBase,
    ModelInitializerBase,
    ModelTrainerBase,
    TrainingInfo,
    model_finder_factory,
    model_initializer_factory,
    model_trainer_factory,
)
from caikit.core.model_management.local_model_initializer import LocalModelInitializer
from caikit.core.module_backends import BackendBase, backend_types
from caikit.core.modules import ModuleBase, ModuleConfig
from sample_lib.modules import SampleModule


# Add mock backend
# This is set in the base test config's load_priority list
class MockBackend(BackendBase):
    backend_type = "MOCK"

    def __init__(self, config=...) -> None:
        super().__init__(config)
        self._started = False

    def start(self):
        self._started = True

    def register_config(self, config):
        self.config = {**config, **self.config}

    def stop(self):
        self._started = False


backend_types.register_backend_type(MockBackend)

# Add a new model finder that tests can use
class TestFinder(ModelFinderBase):
    name = "TestFinder"
    __test__ = False

    def __init__(self, config, instance_name):
        # Fill in expected fields for SampleModule
        config.setdefault("module_id", SampleModule.MODULE_ID)
        config.setdefault("train", {}).setdefault("batch_size", 1)
        config.setdefault("train", {}).setdefault("learning_rate", 0.1)
        self._fail_to_find = config.fail_to_find
        self._raise_on_find = config.raise_on_find
        self._config = config
        self._local_finder = model_finder_factory.construct({"type": "LOCAL"})
        self._instance_name = instance_name

    def find_model(self, model_path, *args, **kwargs):
        if self._raise_on_find:
            raise RuntimeError("You told me to")
        if self._fail_to_find:
            return None
        if os.path.exists(model_path):
            return self._local_finder.find_model(model_path, *args, **kwargs)
        return ModuleConfig(self._config)


model_finder_factory.register(TestFinder)


# Add a new model initializer that tests can use
class TestInitializer(ModelInitializerBase):
    name = "TestInitializer"
    __test__ = False

    def __init__(self, config, instance_name):
        self.config = config
        self.instance_name = instance_name
        self.loaded_models = []
        self.local_initializer = model_initializer_factory.construct({"type": "LOCAL"})

    def init(self, model_config: ModuleConfig, *args, **kwargs) -> Optional[ModuleBase]:
        # allow config.model_type to control whether this loader barfs
        if "model_type" in self.config and "model_type" in kwargs:
            if self.config["model_type"] != kwargs["model_type"]:
                # Don't load in this loader
                return None
        # use the "Local" loader to actually load the model
        model = self.local_initializer.init(model_config)
        self.loaded_models.append(model)
        return model


model_initializer_factory.register(TestInitializer)


# Add a new simple trainer for tests to use
class TestTrainer(ModelTrainerBase):
    name = "TestTrainer"
    __test__ = False

    def __init__(self, config, instance_name):
        self.instance_name = instance_name
        self.canned_status = config.get("canned_status", TrainingStatus.RUNNING)
        self._futures = {}

    class TestModelFuture(ModelTrainerBase.ModelFutureBase):
        __test__ = False

        def __init__(self, parent, trained_model, save_path, save_with_id):
            super().__init__(
                trainer_name=parent.instance_name,
                training_id=str(uuid.uuid4()),
                save_path=save_path,
                save_with_id=save_with_id,
            )
            self._parent = parent
            self._trained_model = trained_model
            self._canceled = False
            self._completed = False

        def get_info(self):
            if self._completed:
                return TrainingInfo(status=TrainingStatus.COMPLETED)
            if self._canceled:
                return TrainingInfo(status=TrainingStatus.CANCELED)
            return TrainingInfo(status=self._parent.canned_status)

        def cancel(self):
            self._canceled = True

        def wait(self):
            self._completed = True
            if self.save_path:
                self._trained_model.save(self.save_path)

        def load(self):
            return self._trained_model

    def train(
        self,
        module_class: Type[ModuleBase],
        *args,
        save_path: Optional[str] = None,
        save_with_id: bool = False,
        **kwargs,
    ):
        trained_model = module_class.train(*args, **kwargs)
        future = self.TestModelFuture(self, trained_model, save_path, save_with_id)
        self._futures[future.id] = future
        return future

    def get_model_future(self, training_id: str) -> ModelTrainerBase.ModelFutureBase:
        if training_id not in self._futures:
            raise ValueError(f"Unknown training id: {training_id}")
        return self._futures[training_id]


model_trainer_factory.register(TestTrainer)


def configured_backends():
    local_initializers = [
        loader
        for loader in MODEL_MANAGER._initializers.values()
        if isinstance(loader, LocalModelInitializer)
    ]
    return [backend for loader in local_initializers for backend in loader._backends]



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from unittest.mock import patch

# Local
from caikit.core.toolkit.compatibility import (
    is_importable_on_platform,
    unsupported_platforms,
)

# Unit Test Infrastructure
from tests.base import TestCaseBase


class TestCompatibilityUtilities(TestCaseBase):
    @classmethod
    def setUpClass(cls):
        cls.echo_input = "hello world!"

    @staticmethod
    def build_unsupported_echo(platform_types):
        @unsupported_platforms(platform_types)
        def echo(some_arg):
            """The function that we are testing our decorator against."""
            return some_arg

        return echo

    @patch("sys.platform", "linux")
    def test_os_compatibility_happy_platform(self):
        """Ensure if we run a supported platform with one unsupported, we can run normally"""
        # Build an echo that explodes on Darwin [OS X], test against running Linux
        deco_echo = TestCompatibilityUtilities.build_unsupported_echo("darwin")
        # Ensures that args / kwargs forward appropriately
        echo_result = deco_echo(self.echo_input)
        self.assertEqual(echo_result, self.echo_input)

    @patch("sys.platform", "linux")
    def test_os_compatibility_happy_platforms(self):
        """Ensure if we run a supported platform with multiple unsupported, we can run normally"""
        deco_echo = TestCompatibilityUtilities.build_unsupported_echo(
            ["win32", "darwin"]
        )
        echo_result = deco_echo(self.echo_input)
        self.assertEqual(echo_result, self.echo_input)

    @patch("sys.platform", "linux")
    def test_os_compatibility_sad_platform_string(self):
        """Ensure if we run the only unspported platform, we explode before calling"""
        # Build an echo that explodes on Linux, test against running Linux
        deco_echo = TestCompatibilityUtilities.build_unsupported_echo("linux")
        with self.assertRaises(NotImplementedError):
            deco_echo(self.echo_input)

    @patch("sys.platform", "linux")
    def test_os_compatibility_sad_platforms(self):
        """Ensure if we run one of many unspported platforms, we explode before calling"""
        # Build an echo that explodes on Linux & Darwin [OS X], test against running Linux
        deco_echo = TestCompatibilityUtilities.build_unsupported_echo(
            ["linux", "darwin"]
        )
        with self.assertRaises(NotImplementedError):
            deco_echo(self.echo_input)

    def test_os_compatibility_bad_decorator_arg_types(self):
        """Ensure if the decorator type is wrong, we explode when the decorator runs"""
        with self.assertRaises(TypeError):
            TestCompatibilityUtilities.build_unsupported_echo(100)

    def test_os_compatibility_bad_decorator_arg_inner_types(self):
        """Ensure if the decorator inner type is wrong, we explode when the decorator runs"""
        with self.assertRaises(TypeError):
            TestCompatibilityUtilities.build_unsupported_echo([100])

    ###############################################################################################
    #                   Tests for OS-based conditional error handling on imports                  #
    ###############################################################################################
    @patch("importlib.import_module")
    @patch("sys.platform", "darwin")
    def test_is_importable_on_happy_os(self, import_module_mock):
        """Ensure that if we can import on a happy platform, we return True"""
        self.assertTrue(
            is_importable_on_platform(
                ".foo", "bar", ImportError, "linux", "Some message"
            )
        )

    @patch("importlib.import_module")
    @patch("sys.platform", "darwin")
    def test_is_importable_happy_on_sad_os(self, import_module_mock):
        """Ensure that we return True if we can import on a platform with special error handling"""
        self.assertTrue(
            is_importable_on_platform(
                ".foo", "bar", ImportError, "darwin", "Some message"
            )
        )

    @patch("importlib.import_module")
    @patch("sys.platform", "darwin")
    def test_unhandled_import_failure_raises_with_sad_os(self, import_module_mock):
        """Ensure that we still raise if an error type we aren't expecting raises on import"""
        import_module_mock.side_effect = ValueError("Boom!")
        with self.assertRaises(ValueError):
            # NOTE: Import only runs through this function if it's a platform we watch, so we
            # need to match our patched value, which in this case is Darwin, to raise; otherwise
            # we would just return True outright and go on to import after the conditional.
            is_importable_on_platform(
                ".foo", "bar", ImportError, "darwin", "Some message"
            )

    @patch("importlib.import_module")
    @patch("sys.platform", "linux")
    def test_unhandled_import_failure_raises_with_happy_os(self, import_module_mock):
        """Ensure that import doesn't run if this is not a watched platform"""
        import_module_mock.side_effect = ValueError("Boom")
        self.assertTrue(
            is_importable_on_platform(
                ".foo", "bar", ImportError, "darwin", "Some message"
            )
        )

    @patch("importlib.import_module")
    @patch("sys.platform", "darwin")
    def test_is_importable_sad_on_handled_err_type_on_sad_os(self, import_module_mock):
        """Ensure we don't import or raise if something is watched and not importable"""
        import_module_mock.side_effect = ImportError("This one is handled!")
        self.assertFalse(
            is_importable_on_platform(
                ".foo", "bar", ImportError, "darwin", "Some message"
            )
        )

    @patch("importlib.import_module")
    @patch("sys.platform", "darwin")
    def test_is_importable_handles_multiple_os_targets(self, import_module_mock):
        """Ensure we don't import or raise if something is watched and not importable"""
        import_module_mock.side_effect = ImportError("This one is handled!")
        self.assertFalse(
            is_importable_on_platform(
                ".foo", "bar", ImportError, ["darwin", "linux"], "Some message"
            )
        )

    @patch("importlib.import_module")
    @patch("sys.platform", "darwin")
    def test_is_importable_validates_outer_arg_types(self, import_module_mock):
        """Ensure that if we replace any arg with a bad type, we raise a TypeError before import"""
        import_module_mock.side_effect = Exception("This will never happen!")
        happy_args = [".foo", "bar", ImportError, "darwin", "Some message"]
        for arg_idx in range(len(happy_args)):
            sad_args = happy_args.copy()
            sad_args[arg_idx] = None
            with self.assertRaises(TypeError):
                is_importable_on_platform(*sad_args)

    @patch("importlib.import_module")
    @patch("sys.platform", "darwin")
    def test_is_importable_validates_iterable_arg_types(self, import_module_mock):
        """Ensure we replace our iterable arg with bad types, our handling is still happy"""
        import_module_mock.side_effect = Exception("This will never happen!")
        with self.assertRaises(TypeError):
            is_importable_on_platform(
                ".foo", "bar", ImportError, [1, 2, 3, 4], "Some message"
            )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""This module contains the python protobufs messages defined in the protobufs interfaces.
These are used as a backend and for serialization of caikit.core objects but should generally
be used only internally or by those intending to serialze data structures. Otherwise,
you are probably interested in the classes found directly in data_model.
"""

# Standard
import importlib
import os


def import_protobufs(proto_dir, package_base_name, current_globals):
    """Add protobufs definitions to the data model so that we can create custom models.
    To do this, we need to provide the path to the proto directory, the base package
    name, and the globals dict for the package being initialized. Usually this will be called
    in your __init__.py within your protobufs package, and look a lot like the following.

    import os

    # Get the import helper from the core
    from caikit.core.data_model.protobufs import import_protobufs
    proto_dir = os.path.dirname(os.path.realpath(__file__))

    # Import all probobufs as extensions to the core
    import_protobufs(proto_dir, __name__, globals())

    While we could do something like this with introspection, things (unfortunately) don't play
    nice with inspecting a wheel whose contents have been compiled to bytecode. :(

    Args:
        proto_dir (str): Path to the proto directory, i.e., the directory that
            you __init__ protobufs file is in.
        package_base_name (str): full name of your package, e.g., __name__ from
            the __init__ protobufs file.
        current_globals (dict): global dictionary from your protobufs package
            __init__ file.
    """

    # look for *_pb2.py files in proto_dir, we will consider these to be our protobufs files
    module_names = [
        filename.rstrip(".py")
        for filename in os.listdir(proto_dir)
        if filename.endswith("_pb2.py")
    ]

    # if there are no modules discovered, fallback to looking for .pyc files
    # this is necessary for binary-only releases
    if not module_names:
        module_names = [
            filename.rstrip(".pyc")
            for filename in os.listdir(proto_dir)
            if filename.endswith("_pb2.pyc")
        ]

    # dynamically load all protobufs files as relative modules
    all_modules = [
        importlib.import_module("." + module_name, package=package_base_name)
        for module_name in module_names
    ]

    # name of protobuf package to use, we ignore anything not in caikit_data_model for now
    _package_name = "caikit_data_model"

    # add all protobufs messages to current module and to the core's data_model
    all_enum_names = []
    for module in all_modules:
        if module.DESCRIPTOR.package.startswith(_package_name):
            for message_name in module.DESCRIPTOR.message_types_by_name.keys():
                message_val = getattr(module, message_name)
                current_globals[message_name] = message_val
                globals()[message_name] = message_val
            for enum_name in module.DESCRIPTOR.enum_types_by_name.keys():
                enum_val = getattr(module, enum_name)
                current_globals[enum_name] = enum_val
                globals()[enum_name] = enum_val
                all_enum_names.append(enum_name)
    current_globals["all_enum_names"] = all_enum_names



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from importlib.metadata import version
from typing import Optional, Type, Union
import traceback

# Third Party
from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.message import Message as ProtoMessageType
from grpc import ServicerContext, StatusCode

# First Party
import alog

# Local
from caikit import get_config
from caikit.core import MODEL_MANAGER, ModuleBase
from caikit.interfaces.common.data_model.stream_sources import S3Path
from caikit.interfaces.runtime.data_model import TrainingJob
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.service_factory import ServicePackage
from caikit.runtime.service_generation.rpcs import ModuleClassTrainRPC
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.utils.import_util import clean_lib_names, get_data_model
from caikit.runtime.utils.servicer_util import (
    build_caikit_library_request_dict,
    validate_data_model,
)
import caikit.core

log = alog.use_channel("GT-SERVICR-I")
error = caikit.core.exceptions.error_handler.get(log)

# Protobuf non primitives
# Ref: https://developers.google.com/protocol-buffers/docs/reference/cpp/google.protobuf.descriptor
NON_PRIMITIVE_TYPES = [FieldDescriptor.TYPE_MESSAGE, FieldDescriptor.TYPE_ENUM]

# pylint: disable=too-many-instance-attributes
class GlobalTrainServicer:
    """Something something about the train servicer"""

    def __init__(self, training_service: ServicePackage):
        self._training_service = training_service
        self._model_manager = ModelManager.get_instance()

        # TODO: think about if we really want to do this here:
        self.cdm = get_data_model()

        # Validate that the Caikit Library CDM is compatible with our service descriptor
        validate_data_model(self._training_service.descriptor)
        log.info("<RUN76773777I>", "Validated Caikit Library CDM successfully")

        # TODO: support multiple libs? `caikit_config.libraries` dict
        # Or grab the `libraries` off of the `training_service` instead of config here?
        # Duplicate code in global_train_servicer
        # pylint: disable=duplicate-code
        self.library = clean_lib_names(get_config().runtime.library)[0]
        try:
            lib_version = version(self.library)
        except Exception:  # pylint: disable=broad-exception-caught
            lib_version = "unknown"

        log.info(
            "<RUN76884779I>",
            "Constructed train service for library: %s, version: %s",
            self.library,
            lib_version,
        )
        super()

    @property
    def training_output_dir(self) -> str:
        return get_config().runtime.training.output_dir

    @property
    def save_with_id(self) -> str:
        return get_config().runtime.training.save_with_id

    def Train(
        self,
        request: ProtoMessageType,
        context: ServicerContext,
        *_,
        **__,
    ) -> TrainingJob:
        """Global train RPC -- Mocks the invocation of a Caikit Library module.train()
        method for a loaded Caikit Library model

        Args:
            request (object): A deserialized RPC request message
            context (ServicerContext): Context object (contains request metadata, etc)

        Returns:
            caikit.interfaces.runtime.data_model.TrainingJob:
                A TrainingJob data model response object
        """
        desc_name = request.DESCRIPTOR.name
        outer_scope_name = "GlobalTrainServicer.Train:%s" % desc_name

        try:
            with alog.ContextLog(log.debug, outer_scope_name):
                module = None
                for mod in caikit.core.registries.module_registry().values():
                    if mod.TASK_CLASS:
                        train_request_for_mod = (
                            ModuleClassTrainRPC.module_class_to_req_name(mod)
                        )
                        if train_request_for_mod == desc_name:
                            module = mod
                            break

                # At this point, if model is still None, we don't know the module this request
                # is for
                if module is None:
                    raise CaikitRuntimeException(
                        StatusCode.INTERNAL,
                        "Global Train not able to parse module for this Train Request",
                    )
                return self.run_training_job(
                    request=request,
                    module=module,
                    training_output_dir=self.training_output_dir,
                    wait=False,
                    context=context,
                ).to_proto()

        except CaikitRuntimeException as e:
            log_dict = {
                "log_code": "<RUN50530380W>",
                "message": e.message,
                "error_id": e.id,
            }
            log.warning({**log_dict, **e.metadata})
            raise e

        # Duplicate code in global_predict_servicer
        # pylint: disable=duplicate-code
        except (TypeError, ValueError) as e:
            log_dict = {
                "log_code": "<RUN72924264W>",
                "message": repr(e),
                "stack_trace": traceback.format_exc(),
            }
            log.warning(log_dict)
            raise CaikitRuntimeException(
                StatusCode.INVALID_ARGUMENT,
                f"Exception raised during training. This may be a problem with your input: {e}",
            ) from e

        except Exception as e:
            log_dict = {
                "log_code": "<RUN24215150W>",
                "message": repr(e),
                "stack_trace": traceback.format_exc(),
            }
            log.warning(log_dict)
            raise CaikitRuntimeException(
                StatusCode.INTERNAL, "Unhandled exception during training"
            ) from e

    def run_training_job(
        self,
        request: ProtoMessageType,
        module: Type[ModuleBase],
        training_output_dir: Optional[str] = None,
        *,
        context: Optional[ServicerContext] = None,
        wait: bool = False,
        **kwargs,
    ) -> TrainingJob:
        """Builds the request dict for calling the train function asynchronously,
        then returns the thread id

        Args:
            request (ProtoMessageType): The message that stimulated this request
            module (Type[ModuleBase]): The module class to train
            training_output_dir (Optional[str]): The base directory where
                the trained model should be saved, falling back to the
                configured output dir if not given.

        Kwargs:
            context (Optional[ServicerContext]): The grpc context for the
                request if called from a grpc handler
            wait (bool): Whether or not to block until the training is complete

        Returns:
            training_job (TrainingJob): The job handle for the training with the
                job's ID and the model's name
        """
        request_data_model = caikit.core.data_model.DataBase.get_class_for_proto(
            request
        ).from_proto(request)

        # Figure out where this model will be saved
        model_path: Union[str, S3Path]
        if request_data_model.output_path:
            # If we got an S3 storage link, just pass that along to the trainer
            model_path: S3Path = request_data_model.output_path
        else:
            # Otherwise, use either:
            # 1. The provided `training_output_dir` here, or
            # 2. The configured `runtime.training.output_dir`
            model_path: str = training_output_dir or self.training_output_dir

        # Build the full set of kwargs for the train call
        kwargs.update(
            {
                "module": module,
                "save_path": model_path,
                "save_with_id": self.save_with_id,
                "model_name": request_data_model.model_name,
                **build_caikit_library_request_dict(
                    request.parameters, module.TRAIN_SIGNATURE
                ),
            }
        )

        # Submit the request to the model manager
        model_future = MODEL_MANAGER.train(**kwargs)

        # if requested, wait for training to complete, thus
        # allowing different servicers to cancel the request
        # in case needed. This does make this call synchronous,
        # but that is the intent of this function, since for async request
        # we have separate function below returning futures.
        # TODO: In future, for the case where we want to execute the training
        # in async manner, we would implement a separate "cancel" / "delete"
        # API which would integrate with different training backends
        # as per their interface requirements.
        if wait:

            # Register the cancellation callback if given a context
            if context is not None:

                # Create a callback to register termination of training
                def rpc_termination_callback():
                    """Cancel the model future if it has not yet completed"""
                    if not model_future.get_info().status.is_terminal:
                        log.warning(
                            "<RUN36361257W>", "Canceling training %s", model_future.id
                        )
                        model_future.cancel()

                # NOTE: callback registration needs to be before waiting for the
                #   future, otherwise request will wait before registering
                #   callback.
                callback_registered = context.add_callback(rpc_termination_callback)
                if not callback_registered:
                    log.warning(
                        "<RUN54118242W>",
                        "Failed to register rpc termination callback, aborting rpc",
                    )
                    raise CaikitRuntimeException(
                        StatusCode.ABORTED,
                        "Could not register RPC callback, call has likely terminated.",
                    )

            with alog.ContextTimer(
                log.debug, "Training %s complete in: ", model_future.id
            ):
                model_future.wait()
                training_info = model_future.get_info()
                if training_info.errors:
                    log.info(
                        "Training %s failed with error: %s. "
                        "Re-raising exception for synchronous response",
                        model_future.id,
                        str(training_info.errors[0]),
                    )
                    raise training_info.errors[0]

        # return TrainingJob object
        return TrainingJob(
            model_name=request.model_name,
            training_id=model_future.id,
        )

    def _load_trained_model(self, model_name: str, model_path: str):
        log.debug("Autoloading trained model %s", model_name)
        self._model_manager.load_model(
            model_id=model_name,
            local_model_path=model_path,
            model_type="standalone",
        )
        return self._model_manager.retrieve_model(model_name)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the model manager, which corrals catalogs to resolve models from names,
and download and load them.
"""
# Standard
from contextlib import contextmanager
import os
import tempfile
import uuid

# Third Party
import pytest

# Local
from caikit.core import LocalBackend
from caikit.core.data_model import DataStream, TrainingStatus
from caikit.core.model_management import ModelFinderBase, model_finder_factory
from caikit.core.modules import ModuleBase, ModuleSaver, module

# Unit Test Infrastructure
from sample_lib.modules.sample_task import SampleModule
from tests.base import TestCaseBase
from tests.conftest import backend_priority, temp_config

# NOTE: We do need to import `reset_backend_types` and `reset_module_distribution_registry` for `reset_globals` to work
from tests.core.helpers import *
import caikit.core


class TestModelManager(TestCaseBase):
    """This test class tries to mock out the catalogs and direct download calls to reduce network
    overhead.
    """

    @classmethod
    def setUpClass(cls) -> None:
        # Test fixtures that can be directly loaded
        cls.model_path = os.path.join(cls.fixtures_dir, "dummy_module")
        # This model has no unique hash set in its config; we use it as a nonsingleton model too
        cls.non_singleton_model_path = cls.model_path
        cls.singleton_model_path = os.path.join(
            cls.fixtures_dir, "dummy_module_singleton"
        )

        cls.resource_path = os.path.join(cls.fixtures_dir, "dummy_resource")

        # Binary buffers of zip archives, for mocking downloads
        cls.module_zip_path = os.path.join(cls.fixtures_dir, "dummy_module.zip")
        with open(cls.module_zip_path, "rb") as f:
            cls.module_archive_buffer = f.read()

    @pytest.fixture
    def global_load_path(self):
        """Set load_path prior to importing caikit.core."""
        # Set test load path
        test_load_path = os.path.join(self.fixtures_dir, "models")

        with temp_config({"load_path": test_load_path}):
            yield

    def test_load_can_return_a_module(self):
        model = caikit.core.load(self.model_path)
        self.assertIsInstance(model, caikit.core.ModuleBase)

    def test_load_can_load_a_module_as_a_singleton(self):
        model1 = caikit.core.load(self.singleton_model_path, load_singleton=True)
        model2 = caikit.core.load(self.singleton_model_path, load_singleton=True)
        assert model1 is model2

    def test_load_can_load_a_module_with_singleton_disabled(self):
        model1 = caikit.core.load(self.singleton_model_path, load_singleton=True)
        model2 = caikit.core.load(self.singleton_model_path, load_singleton=False)
        assert model1 is not model2

    def test_singleton_cache_can_be_cleared(self):
        model = caikit.core.load(self.singleton_model_path, load_singleton=True)
        self.assertGreater(
            len(caikit.core.MODEL_MANAGER.get_singleton_model_cache_info()), 0
        )

        caikit.core.MODEL_MANAGER.clear_singleton_cache()
        self.assertEqual(
            len(caikit.core.MODEL_MANAGER.get_singleton_model_cache_info()), 0
        )

        model2 = caikit.core.load(self.singleton_model_path, load_singleton=True)
        # Pointers should be different, since cache was cleared in between
        self.assertNotEqual(id(model), id(model2))

    def test_load_raises_on_bad_paths(self):
        with self.assertRaises(FileNotFoundError):
            caikit.core.load("bad/path/to/model")

    def test_extract(self):
        with tempfile.TemporaryDirectory() as tempdir:
            extract_path = caikit.core.extract(
                self.module_zip_path, tempdir, force_overwrite=True
            )
            self.assertEqual(extract_path, tempdir)
            self.assertTrue(os.path.isdir(extract_path))
            # load shouldn't barf
            caikit.core.load(extract_path)

    def test_resolve_and_load_can_load_a_model_from_a_path(self):
        """Check that we can resolve paths to module directories"""
        model = caikit.core.resolve_and_load(self.model_path)
        self.assertIsInstance(model, caikit.core.ModuleBase)

    def test_resolve_and_load_returns_back_a_model_reference(self):
        """Check that we can regurgitate things that are already Modules"""
        model = caikit.core.load(self.model_path)
        model = caikit.core.resolve_and_load(model)
        self.assertIsInstance(model, caikit.core.ModuleBase)

    def test_resolve_and_load_rejects_things_that_are_not_strings_or_modules(self):
        """Check that we throw a nice error if the user tries to resolve nonsense"""
        with self.assertRaises(TypeError):
            caikit.core.resolve_and_load(1)
        with self.assertRaises(TypeError):
            caikit.core.resolve_and_load(["test", "not", "a", "module"])
        with self.assertRaises(TypeError):
            caikit.core.resolve_and_load({"this dict is": "not a module"})

    def test_load_model_with_artifacts_from_zip_str(self):
        """Test that we can load a model archive [extracts to temp_dir/...] with artifacts."""
        model = caikit.core.load(self.module_zip_path)
        self.assertIsInstance(model, caikit.core.ModuleBase)

    def test_load_model_with_artifacts_from_bytes(self):
        """Test that we can load a bytes object as a model, even if it has artifacts."""
        model_bytes = caikit.core.load(self.module_zip_path).as_bytes()
        model = caikit.core.load(model_bytes)
        self.assertIsInstance(model, caikit.core.ModuleBase)

    def test_load_model_with_artifacts_from_file_like(self):
        """Test that we can load a file-like object as a model, even if it has artifacts."""
        model_bytesio = caikit.core.load(self.module_zip_path).as_file_like_object()
        model = caikit.core.load(model_bytesio)
        self.assertIsInstance(model, caikit.core.ModuleBase)

    def test_load_model_with_no_nesting(self):
        """Test that we can load a zip even if it unzips directly into the extraction archive."""
        model_path = os.path.join(self.fixtures_dir, "dummy_module_no_nesting.zip")
        model = caikit.core.load(model_path)
        self.assertIsInstance(model, caikit.core.ModuleBase)

    def test_load_invalid_zip_file(self):
        """Test that loading a zip archive not containing a model fails gracefully."""
        model_path = os.path.join(self.fixtures_dir, "invalid.zip")
        with self.assertRaises(FileNotFoundError):
            caikit.core.load(model_path)

    @pytest.mark.usefixtures("global_load_path")
    def test_load_path(self):
        """Test that loading a model from a path defined in the load_path config variable works."""
        model = caikit.core.load("foo")
        self.assertIsInstance(model, caikit.core.ModuleBase)

    def test_import_module_registry(self):
        """Make sure that the module registry can be imported from model_manager"""
        # pylint: disable = import-outside-toplevel,no-name-in-module,unused-import
        from caikit.core.model_manager import module_registry  # isort: skip

    def test_load_model_with_no_module_id(self):
        """Test that we can load a module with the legacy block_id instead of module_id"""
        model_path = os.path.join(self.fixtures_dir, DUMMY_NO_ID_MODEL_NAME)
        model = caikit.core.load(model_path)
        self.assertIsInstance(model, caikit.core.ModuleBase)


# Pytest tests #########################################################

# Setup #########################################################################

DUMMY_MODULE_ID = "foo"

TEST_DATA_PATH = os.path.join("tests", "fixtures")
DUMMY_LOCAL_MODEL_NAME = "dummy_module_foo"
DUMMY_BACKEND_MODEL_NAME = "dummy_module_backend"
DUMMY_NO_ID_MODEL_NAME = "dummy_module_no_id"
CONFIG_FILE_NAME = "config.yml"


def setup_saved_model(mock_backend_class):
    """Fixture to create and cleanup a dummy loadable model"""

    backend_types.register_backend_type(LocalBackend)

    @caikit.core.modules.module(id=DUMMY_MODULE_ID, name="dummy base", version="0.0.1")
    class DummyFoo(caikit.core.ModuleBase):
        @classmethod
        def load(cls, *args, **kwargs):
            return cls()

    # Register backend type
    backend_types.register_backend_type(mock_backend_class)

    @caikit.core.modules.module(base_module=DummyFoo, backend_type=backend_types.MOCK)
    class DummyBar(caikit.core.ModuleBase):
        SUPPORTED_LOAD_BACKENDS = [backend_types.MOCK, backend_types.LOCAL]

        @classmethod
        def load(self, *args, **kwargs):
            return DummyBar()

    return DummyFoo, DummyBar


@caikit.core.modules.module(
    id="non-distributed", name="non distributed mod", version="0.0.1"
)
class NonDistributedModule(caikit.core.ModuleBase):
    @classmethod
    def load(cls, *args, **kwargs):
        return cls()

    def save(self, model_path):
        module_saver = caikit.core.modules.ModuleSaver(
            self,
            model_path=model_path,
        )
        with module_saver:
            pass


@contextmanager
def temp_saved_model(model):
    """Temporarily save the model and yield its path"""
    with tempfile.TemporaryDirectory() as model_path:
        model.save(model_path)
        yield model_path


class NoYamlFinder(ModelFinderBase):
    name = "NOYAML"

    def __init__(self, *_, **__):
        pass

    def find_model(self, model_path):
        """This load function doesn't read from model_path, so it definitely does not read the config.yml file"""
        return ModuleConfig({"module_id": SampleModule.MODULE_ID})


class NoYamlInitializer(ModelInitializerBase):
    name = "NOYAML"

    def __init__(self, *_, **__):
        pass

    def init(self, model_config):
        """This load function doesn't read from model_path, so it definitely does not read the config.yml file"""
        assert model_config.module_id == SampleModule.MODULE_ID
        return SampleModule()


model_finder_factory.register(NoYamlFinder)
model_initializer_factory.register(NoYamlInitializer)


@contextmanager
def setup_test_trainer(config=None):
    factory_blob = {"type": TestTrainer.name}
    if config is not None:
        factory_blob["config"] = config
    with temp_config({"model_management": {"trainers": {"default": factory_blob}}}):
        yield


## Tests #######################################################################


def test_backend_supported_model_load_successfully(reset_globals):
    """Test backend supported model can load successfully"""
    _, DummyBar = setup_saved_model(MockBackend)
    with backend_priority({"type": backend_types.MOCK}):
        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_BACKEND_MODEL_NAME)
        model = caikit.core.load(dummy_model_path)
        assert isinstance(model, DummyBar)


def test_local_model_load_successfully(reset_globals):
    """Test regular / local model can load successfully"""
    DummyFoo, DummyBar = setup_saved_model(MockBackend)
    with backend_priority({"type": backend_types.LOCAL}):
        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_LOCAL_MODEL_NAME)
        model = caikit.core.load(dummy_model_path)
        assert isinstance(model, DummyFoo)
        assert not isinstance(model, DummyBar)


def test_local_model_loaded_backend_successfully(reset_globals):
    """Test if local models can be loaded in specified backend
    if available in SUPPORTED_LOAD_BACKENDS
    """
    _, DummyBar = setup_saved_model(MockBackend)
    with backend_priority({"type": backend_types.MOCK}):
        DummyBar.SUPPORTED_LOAD_BACKENDS.append("LOCAL")
        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_LOCAL_MODEL_NAME)
        model = caikit.core.load(dummy_model_path)
        assert isinstance(model, DummyBar)


def test_backend_model_loaded_as_singleton(reset_globals):
    """Test that backend specific model can be loaded as a singleton"""

    _, DummyBar = setup_saved_model(MockBackend)
    with backend_priority({"type": backend_types.MOCK}):
        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_BACKEND_MODEL_NAME)
        model1 = caikit.core.load(dummy_model_path, load_singleton=True)
        model2 = caikit.core.load(dummy_model_path, load_singleton=True)

        # Pointers should be same
        assert id(model1) == id(model2)


def test_singleton_cache_can_be_cleared(reset_globals):
    """Test that a singleton cache can be cleared"""
    dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_BACKEND_MODEL_NAME)
    # Setup the mock backend
    _, _ = setup_saved_model(MockBackend)
    _ = caikit.core.load(dummy_model_path, load_singleton=True)

    assert len(caikit.core.MODEL_MANAGER.get_singleton_model_cache_info()) > 0

    # Clear cache
    caikit.core.MODEL_MANAGER.clear_singleton_cache()
    assert len(caikit.core.MODEL_MANAGER.get_singleton_model_cache_info()) == 0


def test_singleton_cache_with_different_backend(reset_globals):
    """Test singleton cache doesn't stop different backend models"""

    _, _ = setup_saved_model(MockBackend)
    with backend_priority({"type": backend_types.MOCK}):
        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_BACKEND_MODEL_NAME)
        _ = caikit.core.load(dummy_model_path, load_singleton=True)

        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_LOCAL_MODEL_NAME)
        _ = caikit.core.load(dummy_model_path, load_singleton=True)

        assert len(caikit.core.MODEL_MANAGER.get_singleton_model_cache_info()) == 2


def test_fall_back_to_local(reset_globals):
    """Make sure that if LOCAL is enabled and a given module doesn't have any
    registered backends, the default caikit.core.load is used.
    """
    with temp_saved_model(NonDistributedModule()) as model_path:
        model = caikit.core.load(model_path)

    assert isinstance(model, NonDistributedModule)


def test_load_fails_on_no_supported_backend(reset_globals):
    """Make sure if a given module doesn't have any registered backends,
    loading fails.
    """
    _ = setup_saved_model(MockBackend)
    with backend_priority({"type": backend_types.MOCK}):
        with temp_saved_model(NonDistributedModule()) as model_path:
            with pytest.raises(ValueError):
                caikit.core.load(model_path)


def test_preferred_backend_enabled(reset_globals):
    """Make sure that for a model artifact saved with a local backend can be
    loaded with an enabled non-local backend if given as a preferred_backend.
    """
    _, DummyBar = setup_saved_model(MockBackend)
    with backend_priority({"type": backend_types.MOCK}):
        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_LOCAL_MODEL_NAME)
        model = caikit.core.load(dummy_model_path)
        assert isinstance(model, DummyBar)


def test_module_backend_instance_is_passed_to_load_classmethod(reset_globals):
    """When a module's load function supports the `load_backend` kwarg, the concrete
    instance of the module backend is passed in at load time.
    """
    # re-register the local backend because this test clears out the global registries
    backend_types.register_backend_type(LocalBackend)

    @caikit.core.modules.module(
        id=DUMMY_MODULE_ID, name="backend loader", version="0.0.1"
    )
    class DummyBackendLoadingModule(caikit.core.ModuleBase):
        @classmethod
        def load(cls, model_path: str, load_backend: BackendBase):
            assert load_backend is not None
            return cls()

    dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_LOCAL_MODEL_NAME)
    caikit.core.load(dummy_model_path)


def test_preferred_backend_disabled(reset_globals):
    """Make sure that for a model artifact saved with a local backend loads as
    local even with a preferred_backend when the preferred backend is disabled.
    """
    DummyFoo, DummyBar = setup_saved_model(MockBackend)
    with backend_priority({"type": backend_types.LOCAL}):
        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_LOCAL_MODEL_NAME)
        model = caikit.core.load(dummy_model_path)
        assert isinstance(model, DummyFoo)
        assert not isinstance(model, DummyBar)


def test_non_local_supported_backend(reset_globals):
    """Make sure model artifact saved with as non-local backend loads as
    non-local backend if supported by SUPPORTED_LOAD_BACKENDS
    """
    DummyFoo, _ = setup_saved_model(MockBackend)

    class MockBackend2(BackendBase):
        backend_type = "MOCK2"

        def start(self):
            pass

        def register_config(self, config):
            pass

        def stop(self):
            pass

    backend_types.register_backend_type(MockBackend2)

    @caikit.core.modules.module(base_module=DummyFoo, backend_type=backend_types.MOCK2)
    class DummyBaz(caikit.core.ModuleBase):
        SUPPORTED_LOAD_BACKENDS = [backend_types.MOCK, backend_types.MOCK2]

        @classmethod
        def load(self, *args, **kwargs):
            return DummyBaz()

    with backend_priority({"type": backend_types.MOCK2}):
        dummy_model_path = os.path.join(TEST_DATA_PATH, DUMMY_BACKEND_MODEL_NAME)
        model = caikit.core.load(dummy_model_path)
        assert isinstance(model, DummyBaz)


def test_load_must_return_model():
    """Make sure that the return type of load is checked to be an instance of
    ModuleBase, and will raise TypeError if it is not.
    """

    @caikit.core.module("00110203-baad-beef-0809-0a0b0c0d0e0f", "FunkyModule", "0.0.1")
    class _FunkyModel(SampleModule):
        @classmethod
        def load(cls, model_path):
            return (super().load(model_path), "something else")

    model = _FunkyModel()
    with tempfile.TemporaryDirectory() as tempdir:
        # NOTE: the module will get detected as tests since _FunkyModel is defined here
        model.save(tempdir)
        with pytest.raises(TypeError):
            caikit.core.load(tempdir)


def test_load_with_two_shared_initializers_of_the_same_type(
    good_model_path, reset_globals
):
    """Multiple instances of one shared loader type can be configured"""

    with temp_config(
        {
            "model_management": {
                "finders": {"default": {"type": "LOCAL"}},
                "initializers": {
                    "model-one": {
                        "type": TestInitializer.name,
                        "config": {"model_type": "model one"},
                    },
                    "model-two": {
                        "type": TestInitializer.name,
                        "config": {"model_type": "model two"},
                    },
                    "default": {"type": "LOCAL"},
                },
            }
        }
    ):
        model_one_loader = MODEL_MANAGER.get_initializer("model-one")
        model_two_loader = MODEL_MANAGER.get_initializer("model-two")

        # plain model load should use first loader
        model = caikit.core.load(good_model_path, initializer="model-one")
        assert model_one_loader
        assert model in model_one_loader.loaded_models
        assert model not in model_two_loader.loaded_models

        # model load that fails in the first loader will use the second
        model = caikit.core.load(good_model_path, initializer="model-two")
        assert model not in model_one_loader.loaded_models
        assert model in model_two_loader.loaded_models


def test_load_does_not_read_config_yml_if_loader_does_not_require_it(
    reset_globals, tmp_path
):
    tmpdir = str(tmp_path)
    with open(os.path.join(tmpdir, "config.yml"), "w") as f:
        f.write("{this is not yaml} !!@#$%^")

    with temp_config(
        {
            "model_management": {
                "finders": {
                    "default": {"type": NoYamlFinder.name},
                },
                "initializers": {
                    "default": {"type": NoYamlInitializer.name},
                },
            }
        }
    ):
        model = caikit.core.load(tmpdir)
        assert isinstance(model, SampleModule)


def test_load_loader_finder_by_name(reset_globals):
    """Make sure that non-default loaders and finders can be referenced by name
    when calling load
    """
    with temp_config(
        {
            "model_management": {
                "finders": {
                    "default": {"type": "LOCAL"},
                    "other": {"type": NoYamlFinder.name},
                },
                "initializers": {
                    "default": {"type": "LOCAL"},
                    "other": {"type": NoYamlInitializer.name},
                },
            }
        }
    ):
        model = caikit.core.load("unused", finder="other", initializer="other")
        assert isinstance(model, SampleModule)


def test_load_loader_finder_by_value(reset_globals):
    """Make sure that non-default loaders and finders can be passed by value"""
    with temp_config(
        {
            "model_management": {
                "finders": {
                    "default": {"type": "LOCAL"},
                },
                "initializers": {
                    "default": {"type": "LOCAL"},
                },
            }
        }
    ):
        model = caikit.core.load(
            "unused", finder=NoYamlFinder(), initializer=NoYamlInitializer()
        )
        assert isinstance(model, SampleModule)


def test_initialize_all_components(reset_globals):
    """Make sure that calling initialize_components will proactively create all
    component instances from config
    """
    with temp_config(
        {
            "model_management": {
                "trainers": {
                    "default": {"type": "LOCAL"},
                    "foobar": {"type": "LOCAL"},
                },
                "finders": {
                    "default": {"type": "LOCAL"},
                    "foobar": {"type": "LOCAL"},
                },
                "initializers": {
                    "default": {"type": "LOCAL"},
                    "foobar": {"type": "LOCAL"},
                },
            }
        }
    ):
        assert not MODEL_MANAGER._trainers
        assert not MODEL_MANAGER._finders
        assert not MODEL_MANAGER._initializers
        MODEL_MANAGER.initialize_components()
        assert set(MODEL_MANAGER._trainers.keys()) == {"default", "foobar"}
        assert set(MODEL_MANAGER._finders.keys()) == {"default", "foobar"}
        assert set(MODEL_MANAGER._initializers.keys()) == {"default", "foobar"}


def test_find_without_on_disk_model_no_load_path(good_model_path, reset_globals):
    """Make sure that ephemeral models can be found without existing on disk"""

    with temp_config(
        {
            "model_management": {
                "finders": {"default": {"type": TestFinder.name}},
                "initializers": {"default": {"type": "LOCAL"}},
            }
        }
    ):
        model = caikit.core.load("some_pretend_model")
        assert model
        assert not os.path.exists("some_pretend_model")


def test_find_without_on_disk_model_and_load_path(good_model_path, reset_globals):
    """Make sure that ephemeral models can be found without existing on disk
    when a shared load_path is configured
    """
    with tempfile.TemporaryDirectory() as workdir:
        with temp_config(
            {
                "load_path": workdir,
                "model_management": {
                    "finders": {"default": {"type": TestFinder.name}},
                    "initializers": {"default": {"type": "LOCAL"}},
                },
            }
        ):
            model = caikit.core.load("some_pretend_model")
            assert model
            assert not os.path.exists("some_pretend_model")


def test_load_requires_path_str(reset_globals):
    """Make sure that modules whose load function requires a path string
    explicitly can be loaded
    """
    mod_id = str(uuid.uuid4())

    class CustomException(Exception):
        """Never been seen before!"""

    @module(id=mod_id, name="Needs a path", version="1.2.3")
    class NeedsAPath(ModuleBase):
        def save(self, model_path: str):
            with ModuleSaver(self, model_path=model_path) as module_saver:
                module_saver.update_config({"foo": "bar"})

        @classmethod
        def load(cls, model_path: str, *_, **__):
            if not isinstance(model_path, str) or not os.path.exists(model_path):
                raise CustomException("Yikes!")
            return cls()

    with tempfile.TemporaryDirectory() as workdir:
        model_path = os.path.join(workdir, "test_model")
        NeedsAPath().save(model_path)
        loaded = caikit.load(model_path)
        assert loaded


def test_train_by_module_class(reset_globals):
    """Make sure training can be accessed through the central train function
    with the class directly
    """
    with setup_test_trainer():
        train_future = caikit.train(SampleModule, DataStream.from_iterable([]))
        assert isinstance(train_future, ModelTrainerBase.ModelFutureBase)
        loaded_model = train_future.load()
        assert isinstance(loaded_model, SampleModule)
        found_future = caikit.get_model_future(train_future.id)
        assert found_future is train_future


def test_train_by_module_id(reset_globals):
    """Make sure training can be accessed through the central train function
    with the module_id for the module to train
    """
    with setup_test_trainer():
        train_future = caikit.train(
            SampleModule.MODULE_ID, DataStream.from_iterable([])
        )
        assert isinstance(train_future, ModelTrainerBase.ModelFutureBase)
        loaded_model = train_future.load()
        assert isinstance(loaded_model, SampleModule)
        found_future = caikit.get_model_future(train_future.id)
        assert found_future is train_future


def test_train_with_wait(reset_globals):
    """Make sure calling train with wait correctly waits for completion"""
    with setup_test_trainer():
        # Call without waiting and make sure it's not COMPLETED
        unfinished_train_future = caikit.train(
            SampleModule, DataStream.from_iterable([])
        )
        assert unfinished_train_future.get_info().status == TrainingStatus.RUNNING
        # Call with wait and make sure it is COMPLETED
        finished_train_future = caikit.train(
            SampleModule, DataStream.from_iterable([]), wait=True
        )
        assert finished_train_future.get_info().status == TrainingStatus.COMPLETED


def test_train_with_save_path(reset_globals):
    """Make sure calling train with a save_path correctly saves out the model"""
    with setup_test_trainer():
        with tempfile.TemporaryDirectory() as workdir:
            save_path = os.path.join(workdir, "saved_model")
            train_future = caikit.train(
                SampleModule, DataStream.from_iterable([]), save_path=save_path
            )
            assert train_future.get_info().status == TrainingStatus.RUNNING
            assert not os.path.exists(save_path)
            assert train_future.save_path == save_path
            train_future.wait()
            assert train_future.get_info().status == TrainingStatus.COMPLETED
            assert os.path.exists(save_path)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import os

# Third Party
from transformers import pipeline  # pylint: disable=import-error

# Local
from caikit.core import ModuleBase, ModuleLoader, ModuleSaver, TaskBase, module, task
from text_sentiment.data_model.classification import ClassificationPrediction, ClassInfo


@task(
    required_parameters={"text_input": str},
    output_type=ClassificationPrediction,
)
class HuggingFaceSentimentTask(TaskBase):
    pass


@module(
    "8f72161-c0e4-49b0-8fd0-7587b3017a35",
    "HuggingFaceSentimentModule",
    "0.0.1",
    HuggingFaceSentimentTask,
)
class HuggingFaceSentimentModule(ModuleBase):
    """Class to wrap sentiment analysis pipeline from HuggingFace"""

    def __init__(self, model_path) -> None:
        super().__init__()
        loader = ModuleLoader(model_path)
        config = loader.config
        model = pipeline(model=config.hf_artifact_path, task="sentiment-analysis")
        self.sentiment_pipeline = model

    def run(  # pylint: disable=arguments-differ
        self, text_input: str
    ) -> ClassificationPrediction:
        """Run HF sentiment analysis
        Args:
            text_input: str
        Returns:
            ClassificationPrediction: predicted classes with their confidence score.
        """
        raw_results = self.sentiment_pipeline([text_input])

        class_info = []
        for result in raw_results:
            class_info.append(
                ClassInfo(class_name=result["label"], confidence=result["score"])
            )
        return ClassificationPrediction(class_info)

    @classmethod
    def bootstrap(
        cls, model_path="distilbert-base-uncased-finetuned-sst-2-english"
    ):  # pylint: disable=arguments-differ
        """Load a HuggingFace based caikit model
        Args:
            model_path: str
                Path to HuggingFace model
        Returns:
            HuggingFaceModel
        """
        return cls(model_path)

    def save(self, model_path, **kwargs):  # pylint: disable=arguments-differ
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )

        # Extract object to be saved
        with module_saver:
            # Make the directory to save model artifacts
            rel_path, _ = module_saver.add_dir("hf_model")
            save_path = os.path.join(model_path, rel_path)
            self.sentiment_pipeline.save_pretrained(save_path)
            module_saver.update_config({"hf_artifact_path": rel_path})

    # this is how you load the model, if you have a caikit model
    @classmethod
    def load(cls, model_path):  # pylint: disable=arguments-differ
        """Load a HuggingFace based caikit model
        Args:
            model_path: str
                Path to HuggingFace model
        Returns:
            HuggingFaceModel
        """
        return cls(model_path)



"""
A streaming module for streaming things!

NB: This is here for backwards-compatibility support with ye-olde basic streaming-out tasks
"""

# Local
from ...data_model.sample import SampleInputType, SampleOutputType, StreamingTask
from caikit.core.data_model import DataStream
from caikit.core.modules import ModuleLoader, ModuleSaver
import caikit.core


@caikit.core.module(
    "00110203-0123-0456-0789-0a0b02dd0e0f", "SampleModule", "0.0.1", StreamingTask
)
class StreamingModule(caikit.core.ModuleBase):
    def __init__(self, stream_size=10):
        super().__init__()
        self.stream_size = stream_size

    @classmethod
    def load(cls, model_path, **kwargs):
        loader = ModuleLoader(model_path)
        config = loader.config
        return cls(config["stream_size"])

    def run(self, sample_input: SampleInputType) -> DataStream[SampleOutputType]:
        """
        Args:
            sample_input (sample_lib.data_model.SampleInputType): the input

        Returns:
            caikit.core.data_model.DataStream[sample_lib.data_model.SampleOutputType]: The output
                stream
        """
        list_ = [
            SampleOutputType(f"Hello {sample_input.name}")
            for x in range(self.stream_size)
        ]
        stream = DataStream.from_iterable(list_)
        return stream

    def save(self, model_path):
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with module_saver:
            config_options = {"stream_size": self.stream_size}
            module_saver.update_config(config_options)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data structures for representing images."""

# Third Party
from PIL import Image as PILImage
import numpy as np

# First Party
from py_to_proto.dataclass_to_proto import Annotated, FieldNumber
import alog

# Local
from .backends import ImagePilBackend
from .package import VISION_PACKAGE
from caikit.core import DataObjectBase, dataobject, error_handler

log = alog.use_channel("DATAM")
error = error_handler.get(log)


@dataobject(package=VISION_PACKAGE)
class Image(DataObjectBase):
    """Data model for an image object; this stores the image in the backend as a PIL image, with
    convenience views to interact with the Image as other formats as needed.
    """

    image_data: Annotated[bytes, FieldNumber(1)]

    def __init__(self, *args, **kwargs):
        # For now, we always delegate to the PIL image backend, because it's the only one.
        # In the future, we may have more views etc, but for now, images as PIL Images is
        # the hub format.
        #
        # NOTE: Currently, the args / kwargs here should match up with the proto class, because
        # .from_proto() requires it.
        self._backend = ImagePilBackend(*args, **kwargs)

    def as_numpy(self) -> np.ndarray:
        """Convert the Image data model to a Numpy Array; here, we delegate to the backend.
        Produced images are of type uint8.
        Note that the produced object will have shape (Rows, Cols, Channels).
        Returns:
            np.ndarray
                Numpy array(uint8) representation of this data model object.
        """
        self._check_initialization()
        return self._backend.as_numpy()

    def as_pil(self) -> PILImage.Image:
        """Convert the Image data model to a PIL image. Since we use PIL images as our hub format,
        this is simply returning a handle to the internally stored PIL image.
        Returns:
            PIL.Image
                PIL image representation of this data model object.
        """
        self._check_initialization()
        return self._backend.as_pil()

    def _check_initialization(self):
        """Ensure that we have a _backend; throw an attribute error if we don't. This will always
        be the case if we go through the initializer, but may occur if we call __new__() and use
        this class incorrectly.
        """
        if not isinstance(getattr(self, "_backend", None), ImagePilBackend):
            error(
                "<COR14411171E>",
                AttributeError(
                    "Unable to create view; missing PIL backend initialization"
                ),
            )

    ### Other attributes
    @property
    def rows(self) -> int:
        """Grab the number of rows in the image.
        Returns:
            int
                Number of rows in the underlying image.
        """
        return self.as_pil().height

    @property
    def columns(self) -> int:
        """Grab the number of columns in the image.
        Returns:
            int
                Number of columns in the underlying image.
        """
        return self.as_pil().width

    @property
    def channels(self) -> int:
        """Grab the number of channels in the image.
        Returns:
            int
                Number of channels in the underlying image.
        """
        return len(self.as_pil().getbands())



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A ModelInitializer is responsible for taking an in-memory ModuleConfig and
producing a usable Module instance.

Configuration for ModelInitializers lives under the config as follows:

model_management:
    initializers:
        <initializer name>:
            type: <initializer type name>
            config:
                <config option>: <value>
"""

# Standard
from typing import Optional
import abc

# Local
from ..modules import ModuleBase, ModuleConfig
from ..toolkit.factory import FactoryConstructible


class ModelInitializerBase(FactoryConstructible):
    __doc__ = __doc__

    @abc.abstractmethod
    def init(
        self,
        model_config: ModuleConfig,
        **kwargs,
    ) -> Optional[ModuleBase]:
        """Given a ModelConfig, attempt to initialize it for running

        Args:
            model_config (ModuleConfig): The in-memory model config object for
                the model to be loaded
            **kwargs: All loaders must allow additional kwargs through so that
                specific finders and loaders can support additional optional
                arguments.

        Returns:
            model (Optional[ModuleBase]): The in-memory ModuleBase instance that
                is ready to run if successful. If unable to load, an exception
                may be raised, or None may be returned.
        """



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from datetime import datetime
from typing import List

# First Party
from py_to_proto.dataclass_to_proto import Annotated, FieldNumber
import alog

# Local
from caikit.core.data_model import DataObjectBase, TrainingStatus, dataobject
from caikit.core.toolkit.wip_decorator import Action, WipCategory, work_in_progress

log = alog.use_channel("MDLOPS")

RUNTIME_PACKAGE = "caikit_data_model.runtime"


@work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
@dataobject(RUNTIME_PACKAGE)
class TrainingInfoRequest(DataObjectBase):
    training_id: str


@work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
@dataobject(RUNTIME_PACKAGE)
class TrainingJob(DataObjectBase):
    training_id: str
    model_name: str


@work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
@dataobject(RUNTIME_PACKAGE)
class ModelPointer(DataObjectBase):
    model_id: str


@work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
@dataobject(RUNTIME_PACKAGE)
class TrainingStatusResponse(DataObjectBase):
    training_id: Annotated[str, FieldNumber(1)]
    state: Annotated[TrainingStatus, FieldNumber(2)]
    submission_timestamp: Annotated[datetime, FieldNumber(3)]
    completion_timestamp: Annotated[datetime, FieldNumber(4)]
    reasons: Annotated[List[str], FieldNumber(5)]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobufs-defined services."""
# Third Party
import grpc

# Local
from . import process_pb2 as process__pb2


class ProcessStub(object):
    """A Process running in Model-Train's Dynamic Workflow (DWF) must implement these APIs."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Run = channel.unary_unary(
            "/processproto.Process/Run",
            request_serializer=process__pb2.ProcessRequest.SerializeToString,
            response_deserializer=process__pb2.ProcessResponse.FromString,
        )


class ProcessServicer(object):
    """A Process running in Model-Train's Dynamic Workflow (DWF) must implement these APIs."""

    def Run(self, request, context):
        """Upon receiving a training job, Run() will be called once it is the current process' turn to execute"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_ProcessServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Run": grpc.unary_unary_rpc_method_handler(
            servicer.Run,
            request_deserializer=process__pb2.ProcessRequest.FromString,
            response_serializer=process__pb2.ProcessResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "processproto.Process", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Process(object):
    """A Process running in Model-Train's Dynamic Workflow (DWF) must implement these APIs."""

    @staticmethod
    def Run(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/processproto.Process/Run",
            process__pb2.ProcessRequest.SerializeToString,
            process__pb2.ProcessResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import json
import os
import sys

# First Party
import alog

# Local
from ..core.data_model import render_dataobject_protos
from .service_factory import ServicePackageFactory
from caikit.config.config import get_config
import caikit

log = alog.use_channel("RUNTIME-DUMP-SVC")


def dump_grpc_services(output_dir: str):
    """Utility for rendering the all generated interfaces to proto files"""
    inf_svc = ServicePackageFactory.get_service_package(
        ServicePackageFactory.ServiceType.INFERENCE,
    )
    train_svc = ServicePackageFactory.get_service_package(
        ServicePackageFactory.ServiceType.TRAINING,
    )
    train_mgt_svc = ServicePackageFactory.get_service_package(
        ServicePackageFactory.ServiceType.TRAINING_MANAGEMENT,
    )
    render_dataobject_protos(output_dir)
    inf_svc.service.write_proto_file(output_dir)
    train_svc.service.write_proto_file(output_dir)
    train_mgt_svc.service.write_proto_file(output_dir)


def dump_http_services(output_dir: str):
    """Dump out the openapi.json for the HTTP server"""

    # Import the HTTP components inside the dump function to avoid requiring
    # them when dumping grpc interfaces without the `runtime-http` optional
    # dependencies installed.

    try:
        # Third Party
        from fastapi.testclient import (  # pylint: disable=import-outside-toplevel
            TestClient,
        )

        # Local
        from .http_server import (  # pylint: disable=import-outside-toplevel
            RuntimeHTTPServer,
        )
    except ModuleNotFoundError as e:
        message = (
            "Error: {} - unable to dump http services. Perhaps you missed"
            " installing the http optional dependencies?".format(e)
        )
        log.error("<DMP76165827E>", message)
        sys.exit(1)

    server = RuntimeHTTPServer()
    with TestClient(server.app) as client:
        response = client.get("/openapi.json")
        response.raise_for_status()

        # create output dir if doesn't exist
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        with open(
            os.path.join(output_dir, "openapi.json"), "w", encoding="utf-8"
        ) as handle:
            handle.write(json.dumps(response.json(), indent=2))


if __name__ == "__main__":
    assert len(sys.argv) == 2, f"Usage: {sys.argv[0]} <output_dir>"
    out_dir = sys.argv[1]
    # Set up logging so users can set LOG_LEVEL etc
    caikit.core.toolkit.logging.configure()

    if get_config().runtime.grpc.enabled:
        dump_grpc_services(out_dir)
    if get_config().runtime.http.enabled:
        dump_http_services(out_dir)
